package Fb::Cache;

use 5.006;
use strict;
use warnings;

require Exporter;
our @ISA=qw(Exporter);

sub new {
	my ($package,$options)=@_;
	if ( ! $options ) {
		$options={};
	}
	my $self={
		default_validity => 3600,
		check_interval => 180,
		data => {},
		aliases => {},
		lastcleanup => 0,
	};
	if ( $options->{default_validity} ) {
		$self->{default_validity}=$options->{default_validity};
	}
	if ( $options->{check_interval} ) {
		$self->{check_interval}=$options->{check_interval};
	}
	my $me=bless($self,$package);
}


sub set {
	my ($self,$key,$value,%options)=@_;
	if ( ref($value) eq 'CODE' ) {
		$value=&{$value}();
	}
	$self->remove($key) || return undef;
	if ( defined($value) ) {
		my $timeout=time();
		if ( defined($options{validity}) ) {
			$timeout += $options{validity};
		} else {
			$timeout += $self->{default_validity};
		}
		$self->{data}->{$key}=[$timeout,$value,[]];
	}
	if ( $options{aliases} ) {
		my $aliases=$options{aliases};
		if ( ref($aliases) eq 'CODE' ) {
			$aliases=&{$aliases}($value);
		}
		for my $a (@{$aliases}) {
			$self->{aliases}->{$a}=$self->{data}->{$key};
		}
		$self->{data}->{$key}->[2]=$aliases;
	}
	return $value;
}

sub remove {
	my ($self,$key)=@_;
	if ( $self->{data}->{$key} ) {
		for my $a (@{$self->{data}->{$key}->[2]}) {
			delete $self->{aliases}->{$a};
		}
	}
	delete $self->{data}->{$key};
	return 1;
}

sub removePattern {
	my ($self,$pcre)=@_;
	my @keys=grep { /^$pcre$/ } keys %{$self->{data}};
	my $count=0;
	for my $key (@keys) {
		$count += $self->remove($key);
	}
	return $count;
}

sub clear {
	my $self=$_[0];
	$self->{data}={};
	$self->{aliases}={};
	$self->{lastcleanup}=time();
	return 1;
}

sub keys {
	return [CORE::keys %{$_[0]->{data}}];
}

sub aliases {
	return [CORE::keys %{$_[0]->{aliases}}];
}

sub cleanup {
	my $self=shift;
	my $count=0;
	my $now=time();
	while (my ($key,$item) = each %{$self->{data}}) {
		if ( $item->[0] < $now ) {
			$count += $self->remove($key);
		}
	}
	return $count;
}

sub get {
	my ($self,$key,%options)=@_;

	# Automatic regular cleanup
	my $now=time();
	if ( ! $options{nocleanup} ) {
		if ( ( $now - $self->{lastcleanup} ) > $self->{check_interval} ) {
			$self->cleanup();
		}
	}
	
	my $item=$self->{data}->{$key} || $self->{aliases}->{$key};
	if ( ! defined($item) ) {
		# Not in the cache
	} elsif ( $item->[0] < $now ) {
		# In the cache but timed out
		$self->remove($key);
	} else {
		return $item->[1];
	}
	if ( ! defined($options{create}) ) {
		return undef;
	}
	return $self->set($key,$options{create},%options);
}

1;
