package Fb::Data;

use 5.006;
use strict;
use warnings;
use Data::Dumper;
use Digest;
use Time::Local;
use Carp qw(cluck);

use XML::LibXML;
use XML::LibXML::XPathContext;

my $have_lib_text_diff=0;
eval "use Text::Diff qw(diff);";
if ( $@ eq '' ) {
	$have_lib_text_diff=1;
}

my $have_libdate_calc=0;
eval "use Date::Calc qw(Week_Number Week_of_Year Add_Delta_Days Delta_Days);";
if ( $@ eq '' ) {
	$have_libdate_calc=1;
}

require Exporter;
our @ISA=qw(Exporter);
our @EXPORT_OK=qw(
	hash_dir
	human_readable
	parse_storagesize
	parse_integer
	parse_timeinterval
	check_passwd
	$PASSWD_SPECIAL_CHARS
	$PASSWD_SPECIAL_CHARS_SAFE
	$MAXINT32
	$MAXINT64
	emptyValue
	splitObjects
	randomPassword
	compactArray
	timeStamp
	expandRegex
	hashify
	diffStrings
	align wrapLine limitWords printNewlined outputRecord outputTable
	min max in
	setEquality addToSet removeFromSet setUnion
	listEquality hashEquality
	xml_bool
	bash_quote
	$outmethod
	calcResult
	handleDbField
	leq
	updateEtaStatus
	showEtaStatus
	retireBackups
	getXmlXc
	Dump
);	

# Special characters that are allowed in passwords
our $PASSWD_SPECIAL_CHARS=",.;:_/(){}[]\@%&!+-*~=#?^";
our $PASSWD_SPECIAL_CHARS_SAFE=",.!$%<>";
our $outmethod;
# All valid passwords have to match this regexp
our $VALID_PASSWD_REGEX="[" . quotemeta($PASSWD_SPECIAL_CHARS) . "a-zA-Z0-9]{4,40}";
# To handle virtual infinity
our $MAXINT32=2147483647; # 2^31-1
our $MAXINT64=9223372036854775807; # 2^63-1

sub out {
	return &{$outmethod}(@_);
}

=head2 SOME FUNCTIONS

=over 4

=item $dir=hash_dir($object,$levels)

Returns relative a I<$levels>-level path which is derived from an
I<$object>'s name.
Currently only 2 levels are supported.

=cut

sub hash_dir {
	my ($object,$levels)=@_;
	if ( $levels != 2 ) {
		out('e',"hash_dir(): Only 2 levels hashdirs are supported at this time");
		return undef;
	}
	my $digest=Digest->new("MD5");
	$digest->add($object);
	my $md5=$digest->hexdigest;
	return [substr($md5,0,2),substr($md5,2,2)];
}

=item xx=param_split(...)

Don't know, what this is for...

=cut

sub param_split {
	my ($text,$array_params,$expected_params)=@_;
	my %params=();
	my %array_params=();
	if ( defined($array_params) ) {
		for my $a (@{$array_params}) {
			$array_params{$a}=1;
		}
	}
	if ( defined($expected_params) ) {
		for my $p (@{$expected_params}) {
			if ( defined($array_params{$p}) ) {
				$params{$p}=[];
			} else {
				$params{$p}='';
			}
		}
	}
	for my $p (split /,/,$text) {
		if ( $p =~ /^([^=]+)=(.)$/ ) {
			my $key=$1;my $value=$2;
			if ( defined($array_params{$key}) ) {
				if ( ! defined($params{$key}) ) {
					$params{$key}=[];
				}
				push(@{$params{$key}},$value);
			} else {
				$params{$key}=$value;
			}
		} else {
			$params{$p}=1;
		}
	}
	return \%params;
}


=item $text=human_readable($value,$mode,%options)

Converts value I<$value> in human readable I<$text>. I<$mode> defines
the type of I<$value>. I<$text> will be B<?>, if I<$value> is B<undef>.

Possible values of I<$mode>:

'B<memory>' - A storage size value

'B<timeinterval>' - A time interval (number of seconds)

'B<localtime>' - An epoch value

'B<integer> - a number

Possible I<%options> are:

=over 4

=item B<space>=I<string> Defines the string to be placed between the number and the unit in B<memory> and B<integer> mode.

=item B<limit>=I<value> Defines the maximum number to be displayed before switching to the next higher unit in B<memory> and B<integer> mode.

=back

=cut

sub hrint {
	my ($value,$options)=@_;
	if ( $options->{'radix'} ) {
		return sprintf("%.$options->{'radix'}f",$value);
	}
	return int($value);
}

my @MEMORY=(
	[1,'  '],
	[1024,'ki'],
	[1048576,'Mi'],
	[1073741824,'Gi'],
	[1099511627776,'Ti'],
	[1125899906842624,'Pi'],
	[1152921504606846976,'Ei'],
);
my @INTEGER=(
	[1,' '],
	[1000,'k'],
	[1000000,'M'],
	[1000000000,'G'],
	[1000000000000,'T'],
	[1000000000000000,'P'],
	[1000000000000000000,'E'],
);

sub human_readable {
	my ($value,$mode,%options)=@_;
	if ( ! defined($value) ) {
		return '?';
	}
	if ( defined($options{'hr'}) ) {
		if ( ! $options{'hr'} ) {
			return $value;
		}
	}
	my $limit;
	if ( defined($options{'limit'}) ) {
		$limit=$options{'limit'};
	}
	my $space=' ';
	if ( defined($options{'space'}) ) {
		$space=$options{'space'};
	}
	if ( $mode eq 'memory' ) {
		my $prefix='';
		if ( $value < 0 ) {
			$prefix='-';
			$value=abs($value);
		}
		if ( ! defined($limit) ) {
			$limit=100000;
		}
		for(my $c=0;$c<=$#MEMORY;$c++) {
			if ( $value < ($limit*$MEMORY[$c]->[0]) ) {
				my $suffix=$MEMORY[$c]->[1];
				if ( $c == 0 && $space eq '' ) {
					$suffix='';
				}
				return $prefix . hrint($value/$MEMORY[$c]->[0],\%options) . $space . $suffix;
			}
		}
		# Failsafe for *VERY* big numbers
		return $prefix . hrint($value/$MEMORY[$#MEMORY]->[0],\%options) . $space . $MEMORY[$#MEMORY]->[1];
	} elsif ( $mode eq 'integer' ) {
		if ( ! defined($value) ) {
			return "undef";
		}
		my $prefix='';
		if ( $value < 0 ) {
			$prefix='-';
			$value=abs($value);
		}
		my $mult=1;
		if ( $value < 0 ) {
			$mult=1;$value=abs($value);
		}
		if ( ! defined($limit) ) {
			$limit=10000;
		}
		for(my $c=0;$c<=$#INTEGER;$c++) {
			if ( $value < ($limit*$INTEGER[$c]->[0]) ) {
				my $suffix=$INTEGER[$c]->[1];
				if ( $c == 0 && $space eq '' ) {
					$suffix='';
				}
				return $prefix . hrint($value/$INTEGER[$c]->[0],\%options) . $space . $suffix;
			}
		}
		return $prefix . hrint($mult*$value/$INTEGER[$#INTEGER]->[0],\%options) . $space . $INTEGER[$#INTEGER]->[1];
	} elsif ( $mode eq 'timeinterval' ) {
		if ( ! defined($limit) ) {
			$limit=10;
		}
		my $prefix='';
		if ( $value < 0 ) {
			$prefix="-";
			$value=abs($value);
		}
		my $out;
		if ( $value > $limit*60 ) {
			if ( $value > $limit*60*60 ) {
				if ( $value > $limit*60*60*24 ) {
					if ( $value > $limit*60*60*24*7 ) {
						if ( $value > $limit*60*60*24*365 ) {
							$out=(hrint($value/60/60/24/365,\%options) . "a");
						} else {
							$out=(hrint($value/60/60/24/7,\%options) . "w");
						}
					} else {
						$out=(hrint($value/60/60/24,\%options) . "d");
					}
				} else {
					$out=(hrint($value/60/60,\%options) . "h");
				}
			} else {
				$out=(hrint($value/60,\%options) . "m");
			}
		} else {
			$out=$value . "s";
		}
		return $prefix . $out;
	} elsif ( $mode eq 'localtime' ) {
		my @l=localtime($value);
		my $text=sprintf('%0.2d.%0.2d.%0.4d %0.2d:%0.2d:%0.2d',$l[3],$l[4]+1,$l[5]+1900,$l[2],$l[1],$l[0]);
		return $text;
	} elsif ( $mode eq 'localtime/short' ) {
		my @l=localtime($value);
		my $text=sprintf("%0.4d%0.2d%0.2d-%0.2d%0.2d%0.2d",$l[5]+1900,$l[4]+1,$l[3],$l[2],$l[1],$l[0]);
		return $text;
	} else {
		out('e',"human_readable(): Unknown mode '$mode', internal error");
		return undef;
	}
}


=item $seconds=parse_timeinterval($timestring,%options)

Interprets timestring, returns seconds. I<$timestring> must be a value, followed by 'B<s>', 'B<m>', 'B<h>', 'B<d>' or 'B<w>'.
 
This is the reverse function of

I<$timestring>=human_readable(I<$seconds>,'timeinterval')

=cut

sub parse_timeinterval {
	my ($time,%options)=@_;
	my $defaultunit;
	if ( $options{defaultunit} ) {
		$defaultunit=$options{defaultunit};
	}
	if ( $time =~ /^([0-9]+)([smhdwMa])?$/ ) {
		my $value=$1;
		my $unit=$2;
		if ( ! defined($unit) ) {
			$unit=$defaultunit;
			if ( ! defined($unit) ) {
				out('w',"time string '$time' without a unit, assuming seconds.");
				return int($time);;
			}
		}
		my %units=(
			's' => 1,
			'm' => 60,
			'h' => 60*60,
			'd' => 60*60*24,
			'w' => 60*60*24*7,
			'M' => 60*60*24*30,
			'a' => 60*60*24*365
		);
		return ($value*$units{$unit});
	} else {
		if ( ! $options{test} ) {
			out('w',"Unable to parse time string '$time'");
		}
		return undef;
	}
}


=item $size=parse_storagesize($text{,$size})

Interprets the given storage value I<$text> which might be either
absolute ( a value + either nothing, B<k>, B<M>, ... ) or a percentage
value relative to I<$size>.

Examples:

 parse_storagesize('1k') == 1024
 
 parse_storagesize('10%',1000) == 100

=cut

sub parse_storagesize {
	my ($text,$size)=@_;
	if ( ! defined($text) ) {
		return undef;
	}
	if ( $text =~ /^([0-9]+)([kMGTP\%]?)[bB]?$/ ) {
		my ($value,$ext)=($1,$2);
		if ( $ext eq 'k' ) {
			return ($value*1024);
		} elsif ( $ext eq 'M' ) {
			return ($value*1024*1024);
		} elsif ( $ext eq 'G' ) {
			return ($value*1024*1024*1024);
		} elsif ( $ext eq 'T' ) {
			return ($value*1024*1024*1024*1024);
		} elsif ( $ext eq 'P' ) {
			return ($value*1024*1024*1024*1024*1024);
		} elsif ( $ext eq '%' ) {
			if ( ! defined($size) ) {
				out('e',"Undefined size and you wanted me to calc percentage");
				return undef;
			}
			return int(($size*$value)/100);
		} else {
			return int($value);
		}
	} else {
		out('w',"Unable to parse storage size '$text', returning undef.");
		return undef;
	}
}

=item $size=parse_integer($text{,$size})

Interprets the given interger value I<$text> which might be either
absolute ( a value + either nothing, B<k>, B<M>, ... ) or a percentage
value relative to I<$size>.

Examples:

 parse_storagesize('1k') == 1000
 
 parse_storagesize('10%',1000) == 100

=cut

sub parse_integer {
	my ($text,$size)=@_;
	if ( ! defined($text) ) {
		return undef;
	}
	if ( $text =~ /^([0-9]+)([kMGTP\%]?)[Bb]?$/ ) {
		my ($value,$ext)=($1,$2);
		if ( $ext eq 'k' ) {
			return ($value*1000);
		} elsif ( $ext eq 'M' ) {
			return ($value*1000*1000);
		} elsif ( $ext eq 'G' ) {
			return ($value*1000*1000*1000);
		} elsif ( $ext eq 'T' ) {
			return ($value*1000*1000*1000*1000);
		} elsif ( $ext eq 'P' ) {
			return ($value*1000*1000*1000*1000*1000);
		} elsif ( $ext eq '%' ) {
			if ( ! defined($size) ) {
				out('e',"Undefined size and you wanted me to calc percentage");
				return undef;
			}
			return int(($size*$value)/100);
		} else {
			return int($text);
		}
	} else {
		out('w',"Unable to parse integer '$text', returning undef.");
		return undef;
	}
}



=item $isvalid=check_passwd($password,%options)

Checks, if the password I<$password> contains valid characters only. Possible
I<%options> are:

=over 4

=item B<test> => B<1>

Ommit showing a descriptive error message.

=cut

sub check_passwd {
	my ($passwd,%options)=@_;
	if ( $passwd !~ /^$VALID_PASSWD_REGEX$/ ) {
		if ( ! $options{test} ) {
			out('e',"This password ($passwd) is not valid.\nYou may use letters/numbers an those special chars: $PASSWD_SPECIAL_CHARS");
		}
		return 0;
	}
	return 1;
}

=item $result=emptyValue($value)

Tests, if I<$value> is considered empty.

=cut

sub emptyValue {
	my $value=shift;
	return ( ! defined($value) || ( $value eq '-' ) || ( $value eq '' ) );
}

=item $result=printNewlined($text)

Prints I<$text> to STDOUT. I<$result> == 1, if I<$text> wasn't B<undef>.

=cut

sub printNewlined {
	my $text=shift;
	if ( ! defined($text) ) {
		return 0;
	}
	if ( substr($text,length($text)-1,1) ne "\n" ) {
		$text .= "\n";
	}
	print $text;
	return 1;
}

=item result=splitObjects(\@objects,%options)

This methods tries to get words from I<@objects>. Words are
separated by spaces and/or B<,>. Possible I<%options> are:

=over 4

=item B<format> (defaults to B<array>)

Set format of the result to either B<array>, B<hash>, B<arrayref> or B<hashref>.

=back

=cut

sub splitObjects {
	my ($objects,%options)=@_;
	my $format='array';
	if ( defined($options{'format'}) ) {
		$format=$options{'format'};
	}
	my %objects=();
	my @objects=();
	for my $p (@{$objects}) {
		my @sub=split /[\s,]+/,$p;
		for my $s (@sub) {
			if ( $s ne '' ) {
				if ( ! $objects{$s} ) {
					push(@objects,$s);
				}
				$objects{$s}=1;
			}
		}
	}
	if ( $format eq 'array' ) {
		# This is sorted
		return @objects;
	} elsif ( $format eq 'arrayref' ) {
		return \@objects;
	} elsif ( $format eq 'hash' ) {
		return %objects;
	} elsif ( $format eq 'hashref' ) {
		return \%objects;
	}
}

=item $password=randomPassword(%options)

Returns a randomly generated password with I<$length> characters length.

Possible I<%options> are:

=over 4

=item B<length> Password length (default: 20)

=item B<charset>

Sets the character set of which to choose password characters from. Possible values are:

=over 4

=item B<base64> [a-zA-Z0-9+/]

=item B<default> Base64 with "/" replaced by "-" [a-zA-Z0-9+-]

=item B<salt> Base64 with "+" replaced by "." [a-z0-9/.] for creating SHA salted hashes.


=item B<hex> [0-9A-F]

=back

=back

=cut

sub randomPassword {
	my (%options)=@_;
	my $length=$options{length};
	if ( ! defined($length) ) {
		$length=20;
	}
	my $charset=$options{charset};
	if ( ! defined($charset) ) {
		$charset='base64';	
	}
	if ( $length > 43 ) {
		my $string='';
		while ( length($string) < $length ) {
			my $part=randomPassword(length => 43,charset => $charset);
			if ( ! defined($part) ) {
				return undef;
			}
			$string .= $part;
		}
		return substr($string,0,$length);
	}
	my $fh;
	if ( ! open($fh,"< /dev/urandom") ) {
		out('e',"Unable to get randomness ( open(/dev/urandom) )");
		return undef;
	}
	my $readlength=1+int($length/4*3);
	my $random;
	if ( read($fh,$random,$readlength) != $readlength ) {
		out('e',"Unable to get enough randomness");
		return undef;
	}
	close($fh);
	my $digest=Digest->new("SHA-256");
	$digest->add($random);
	if ( $charset eq 'default' ) {
		my $s=substr($digest->b64digest,0,$length);
		$s =~ s/\//-/g;
		return $s;
	} elsif ( $charset eq 'salt' ) {
		my $s=substr($digest->b64digest,0,$length);
		$s =~ s/\+/./g;
		return $s;
	} elsif ( $charset eq 'base64' ) {
		return substr($digest->b64digest,0,$length);
	} elsif ( $charset eq 'hex' ) {
		return substr($digest->hexdigest,0,$length)
	}
}

=item $text=timeStamp($format,$time)

Returns a timestamp in the given format I<$format>. I<$time> is an epoch value. I<$time> being B<undef>
means, the current time is used. The I<$time> I<$MAXINT64> is consindered infinity
an will make timeStamp() return B<->.

=cut

sub timeStamp {
	my ($mode,$time)=@_;
	if ( ! defined($time) ) {
		$time=time();
	}
	if ( $time eq $MAXINT64 ) {
		return '-';
	}
	my @t=localtime($time);
	if ( ! @t ) {
		return '?';
	}
	if ( $mode eq 'YYYYMMDD-hhmmss' ) {
		return sprintf("%i%02i%02i-%02i%02i%02i",$t[5]+1900,$t[4]+1,$t[3],$t[2],$t[1],$t[0]);
	} elsif ( $mode eq 'YYYY-MM-DD hh:mm:ss' ) {
		return sprintf("%i-%02i-%02i %02i:%02i:%02i",$t[5]+1900,$t[4]+1,$t[3],$t[2],$t[1],$t[0]);
	} elsif ( $mode eq 'YYYYMMDD-hhmm' ) {
		return sprintf("%i%02i%02i-%02i%02i",$t[5]+1900,$t[4]+1,$t[3],$t[2],$t[1]);
	} elsif ( $mode eq 'DD.MM.YYYY' ) {
		return sprintf("%02i.%02i.%04i",$t[3],$t[4]+1,$t[5]+1900);
	} elsif ( $mode eq 'YYYYMMDD' ) {
		return sprintf("%i%02i%02i",$t[5]+1900,$t[4]+1,$t[3]);
	} elsif ( $mode eq 'YYYY-MM-DD' ) {
		return sprintf("%i-%02i-%02i",$t[5]+1900,$t[4]+1,$t[3]);
	} else {
		out('e',"timeStamp(): Invalid mode '$mode'");
		return undef;
	}
}

=item $example=expandRegex($regexp)

Tries to expand the regular expression I<$regexp> to an I<$example> matching the RE.

=cut

sub expandRegex {
	my $re=shift;
	my $example='';
	while ( $re =~ /(?:\[([^\]]+)\]|([0-9a-zA-Z\.]))(\{([0-9]+)(?:,([0-9]))\})?/g ) {
		my $next;
		my $set=$1;my $literal=$2;
		my $min=$3;my $max=$4;
		if ( defined($literal) ) {
			$next = $literal;
		} else {
			my %set=();
			while ( $set =~ /(?:(\S)-(\S)|(\S))/g ) {
				if ( defined($1) ) {
					my @subset=$1 .. $2;
					for my $element (@subset) {
						$set{$element}=1;
					}
				} else {
					$set{$3}=1;
				}
			}
			my $setexpanded='';
			for my $element (sort keys %set) {
				$setexpanded .= $element;
			}
			$next=substr($setexpanded,int(length($setexpanded)/2),1);
		}
		if ( $min ) {
			if ( $max ) {
				$example .= ($next x $max);
			} else {
				$example .= ($next x $min);
			}
		} else {
			$example .= $next;
		}
	}
	return $example;
}

=item %hash=hashify($x)

Examines I<$x> and converts it into an hash using these rules:

=over 4

=item If it's undefined, convert it into an empty ARRAY/HASH B<()>

=item If it's a scalar, convert it into I<($x => 1)>.

=item If it's a array reference, convert it into a hash with every array element being a key with value B<1> in the new hash.

=item If it's a hash reference, make a hash with every former key being assigned the value B<1>.

=back

This method is used for parameter processing in perl functions.

=cut

sub hashify {
	my $x=shift;
	if ( ! defined($x) ) {
		return ();
	}
	my %hash=();
	if ( ref($x) eq '' ) {
		$hash{$x}=1;
	} elsif ( ref($x) eq 'HASH' ) {
		%hash=%{$x};
	} elsif ( ref($x) eq 'ARRAY' ) {
		for my $tag (@{$x}) {
			$hash{$tag}=1;
		}
	}
	return %hash;
}

=item $result=diffStrings($string1,$string2)

Compares two strings and prints their differences to STDOUT.

I<$result> == 1, only if the strings are equal.

Both strings can be given as arrayrefs of strings. They'll be joined
before comparision. Possible I<%options> are:

=over 4

=item B<getdiff> => I<\$string>

Returns the differences as string instead of printing it to STDOUT.

=back

=cut

sub diffStrings {
	my ($s1,$s2,%options)=@_;
	if ( ! $have_lib_text_diff ) {
		out('w',"libtext-diff-perl is not installed. Unable to show differences");
		return 0;
	}
	my $s1_=$s1;my $s2_=$s2;
	if ( ref($s1) eq 'ARRAY' ) {
		$s1_=join("\n",@{$s1});
		if ( length($s1_) ) {
			$s1_ .= "\n";
		}
	}
	if ( ref($s2) eq 'ARRAY' ) {
		$s2_=join("\n",@{$s2});
		if ( length($s2_) ) {
			$s2_ .= "\n";
		}
	}
	my $diff=diff(\$s1_,\$s2_,{STYLE => "OldStyle"});
	my @lines=grep { /^[<>]/ } split /\n/,$diff;
	if ( @lines ) {
		my $out=join("\n",@lines) . "\n";
		if ( defined($options{getdiff}) ) {
			${$options{getdiff}}=$out;
		} else {
			print $out;
		}
		return 0;
	}
	if ( defined($options{getdiff}) ) {
		${$options{getdiff}}='';
	}
	return 1;
}

=item $text=wrapLine($rawtext,$maxlength,%options)

Makes a I<$text> from the words in I<$rawtext>. Lines in I<$text> will not
exceed I<maxlength> characters - except when words are longer than that.
Possible I<%options> are:

=over 4

=item B<abfprefix>=I<string>

Defines a string to be appended to all but the first line.

=back

=cut

sub wrapLine {
	my ($array,$maxlength,%options)=@_;
	if ( ! defined($array) ) {
		cluck("Undefined array in wrapLine()");
	}
	if ( ref($array) eq '' ) {
		$array=[split /\s+/,$array];
	}
	my $prefix='';my $virtualadd=0;my $text='';my $line='';
	my $first=1;
	if ( defined($options{abfprefix}) ) {
		$prefix=$options{abfprefix};
		$virtualadd=length($prefix);
		if ( defined($options{fprefix}) ) {
			$line=$options{fprefix};
			$virtualadd=0;
		}
	} elsif ( defined($options{prefix}) ) {
		$prefix=$options{prefix};
		$line=$prefix;
	}
	my $lineempty=1;
	for my $word (@{$array}) {
		if ( ! defined($word) ) { $word='(undef)'; }
		if ( ! $lineempty && ( length($line) + length($word) + 1 - $first + $virtualadd ) > $maxlength ) {
			$text .= $line . "\n";
			$line=$prefix . $word;
			$lineempty=0;
			$virtualadd=0;
		} elsif ( $first ) {
			if ( ! $virtualadd ) {
				$line = "$prefix$word";
			} else {
				$line = $word;
			}
			$first=0;
			$lineempty=0;
		} else {
			$line .= " " . $word;
			$first=0;
			$lineempty=0;
		}
	}
	if ( $line ne '' || ! @{$array} ) {
		$text .= $line ."\n";
	}
	return $text;
}

=item $text=limitWords(\@list,$max)

Limits the maximum number of entries in I<@list> to be put into the
space-separated I<$text> to I<$max>. If the maximum is exceeded,
only the number of entries in []-brackets is put into the I<$text>.

=cut

sub limitWords {
	my ($list,$max)=@_;
	my @list;
	if ( ref($list) eq 'ARRAY' ) {
		@list=@{$list};
	} else {
		@list=keys %{$list};
	}
	if ( scalar(@list) > $max ) {
		return "[" . scalar(@list) . "]";
	} else {
		return join(' ',sort @list);
	}
}

sub addToSet {
	my ($set,$tba)=@_;
	for my $e (keys %{$tba}) {
		$set->{$e}=$tba->{$e};
	}
}

sub removeFromSet {
	my ($set,$tbr)=@_;
	for my $e (keys %{$tbr}) {
		delete ($set->{$e});
	}
}

=item $result=setEquality(\%hash1,\%hash2)

Compares two hashes' keys.

I<$result> == 1, only if the hashes' keys are equal.

=cut

sub setEquality {
	my ($seta,$setb)=@_;
	for my $k (keys %{$seta}) {
		if ( ! exists $setb->{$k} ) {
			return 0;
		}
	}
	for my $k (keys %{$setb}) {
		if ( ! exists $seta->{$k} ) {
			return 0;
		}
	}
	return 1;
}

=item $max=max(@list)

Returns the biggest item (I<$max>) from I<@list> which is compared numerically.

=cut

sub max {
	if ( scalar(@_) == 0 ) {
		return undef;
	}
	my $max=shift @_;
	for my $x (@_) {
		if ( $max < $x ) {
			$max=$x;
		}
	}
	return $max;
}

=item $min=min(@list)

Returns the smallest item (I<$min>) from I<@list> which is compared numerically.

=cut

sub min {
	my $min=shift @_;
	for my $x (@_) {
		if ( $min > $x ) {
			$min=$x;
		}
	}
	return $min;
}
                 
=item $result=in(\@list,$test)

Tests, if the given I<$test> string is an element of the I<@list>.
$result == 1, if the string is an element of the list, otherwise 0.

=cut
                               
sub in {
	my ($list,$test)=@_;
	for my $element (@{$list}) {
		if ( $test eq $element ) {
			return 1;
		}
	}
	return 0;
}

sub setUnion {
	my %hash=();
	for my $set (@_) {
		for my $e (@{$set}) {
			$hash{$e}=1;
		}
	}
	return keys %hash;
}


=item $result=listEquality(\@a,\@b);

Tests, if the given arrays contain equal (according to string 'eq') elements.
B<undef> elements on the same positions in I<@a> and I<@b> count as equal.

=cut

sub listEquality {
	my ($a,$b)=@_;
	if ( @{$a} != @{$b} ) {
		return 0;
	}
	for(my $n=0;$n<@{$a};$n++) {
		my $_a=$a->[$n];my $_b=$b->[$n];
		if ( ! defined($_a) ) {
			if ( ! defined($_b) ) {
				next;
			} else {
				return 0;
			}
		}
		if ( ! defined($_b) ) {
			return 0;
		}
		if ( $_a ne $_b ) {
			return 0;
		}
	}
	return 1;
}

=item $result=leq(@list)

Returns I<$result> == B<1>, if all entries of the list are identical
which is one of three cases:

=over 4

=item All item can be string compared and are equal

=item All items are B<undef>

=item The list is empty

=back

B<0> is returned in any other case.

=cut

sub leq {
	my $e1=shift;
	if ( defined($e1) ) {
		for my $e (@_) {
			if ( ! defined($e) || $e ne $e1 ) {
				return 0;
			}
		}
	} else {
		for my $e (@_) {
			if ( defined($e) ) {
				return 0;
			}
		}
	}
	return 1;
}


=item $result=hashEquality(\%a,\%b);

Returns B<1>, only if both hashes contain the same set of keys
associated to same (string comparable) values.

=cut

sub hashEquality {
	my ($a,$b)=@_;
	if ( scalar(keys %{$a}) != scalar(keys %{$b}) ) {
		return 0;
	}
	for my $k (keys %{$a}) {
		if ( ! defined($a->{$k}) and ! defined($b->{$k}) ) {
			next;
		}
		if ( ! exists($b->{$k}) ) {
			return 0;
		}
		if ( $a->{$k} ne $b->{$k} ) {
			return 0;
		}
	}
	return 1;
}


sub xml_bool {
	my $value=shift;
	if ( $value =~ /^true|1$/i ) {
		return 1;
	} elsif ( $value =~ /^false|1$/i ) {
		return 0;
	}
	out('e',"Unable to interpret xml bool '$value'");
	return undef;
}

=item $textq=bash_quote($text)

Returns I<$text> quoted as a single parameter for a shell command.  Target
shell is bash. Empty I<$text>s will make a single parameter, too.
B<undef> however will return an empty string.
It is to be used like this:

$command=join(' ',map { bash_quote($_) } @parameters);
 
=cut

sub bash_quote {
	my $text=shift;
	if ( ! defined($text) ) {
		return "";
	}
	if ( $text eq '' ) {
		return '""';
	}
#	$text =~ s/([\\"\!\|\$\`\ \;\&<>\(\)\'\[\]\{\}\~\*#,%\?])/\\$1/g;
	# We try to cover more special characters
	$text =~ s/([^\w\n])/\\$1/g;
	$text =~ s/\n/\$'\\n'/g;
	return $text;
}

=item $text=align($string,$a)

Aligns a string I<$s> inside a text I<$text> according the alignment
parameter I<$a>. I<$a> < 0 means "right aligned", I<$a> > 0 means "left
aligned". I<$a> == 0 will make align() return the unmodified I<$string>.

=cut

sub align {
	my ($s,$a)=@_;
	if ( ! defined($s) ) {
		return undef;
	}
	my $l=length($s);
	if ( $a > 0 ) {
		if ( $l < $a ) {
			return ($s . (" " x ($a-$l)));
		} elsif ( $l > $a )  {
			return substr($s,0,$a);
		}
	} elsif ( $a < 0 ) {
		$a=abs($a);
		if ( $l < $a ) {
			return ((" " x ($a-$l)) . $s);
		} else {
			return $s;
		}
	}
	return $s;
}

=item $return=calcResult(\$mainresult,$newresult)

This will take the result of a subfunction (I<$newresult>) into account
when maintaining a global I<$mainresult> variable that reflects the
success of a bigger set of function calls.

The idea is to carry the success in bit 0 and the fact if something was
physically changed in bit 1.

I<$return>ed value is bit 0 of $mainresult.

=cut

sub calcResult {
	my ($mainresult,$newresult)=@_;
	${$mainresult}=
		( ${$mainresult} | $newresult ) & 2 +
		( ${$mainresult} & 1 & $newresult );
	return ${$mainresult} & 1;
}

=item $text=outputRecord(\@fields,%options)

Converts an array of I<@fields> given as [I<$key>,I<$value>,%fieldoptions] pairs
into something human readable. I<$value> can be either a string, a arrayref (of
non-separatable words) or a hashref (which are shown as key/value pairs
separated by " - ").

Possible I<%options> are:

=over 4

=item B<prefix> (Default: B<"\t">)

Default prefix for fields. Good choices are B<''> or B<"\t">.

=item B<tablength> (Default: B<3>)

Length of a single tab.

=item B<maxlength> (Default: B<80>)

Width of the terminal. Single lines in I<$text> will never exceed
this.

=item B<keysuffix> (Default: empty string)

String to append to every key (i.e. B<:>).

=item B<description> (Default: B<0>)

Shows the description of a key in addition to the value ( = B<1> )
or instead of the value ( = B<2> ).
 
=back

=cut

sub outputRecord {
	my ($data,%options)=@_;
	if ( ref($data) eq 'HASH' ) {
		my @data;
		for my $key (sort keys %{$data}) {
			push(@data,[$key,$data->{$key}]);
		}
		$data=\@data;
	}
	my $keysuffix='';
	if ( $options{keysuffix} ) {
		$keysuffix=$options{keysuffix};
	}
	my $prefix="\t";
	if ( defined($options{prefix}) ) {
		$prefix=$options{prefix};
	}
	my $tablength=8;
	if ( defined($options{tablength}) ) {
		$tablength=$options{tablength};
	}
	my $maxlength=80;
	if ( defined($options{maxlength}) ) {
		$maxlength=$options{maxlength};
	}

	# Calculate length of the prefix taking tablengths into account
	my $prefix_=$prefix;
	my $tab_=' ' x $tablength;
	$prefix_ =~ s/\t/$tab_/g;
	my $prefixlength=length($prefix_);
	my $emptyprefix=(" " x ($prefixlength+2));

	# Calculate maximum key length
	my $maxkl=max(map { length($_->[0]) } @{$data});
	$maxkl += length($keysuffix);
	
	# Spaces from the beginning of the line until the value should start
	my $inckeyprefix=$prefix . " " x ($maxkl+1);
	
	my $text='';
	for my $field (@{$data}) {
		my @field=@{$field};
		my $key=$field[0];
		my $value=$field[1];
		$text .= $prefix . align($key . $keysuffix,$maxkl) . " ";

		# Output a field description. This is to keep documentation near the code
		my %fieldoptions=@field[2..$#field];
		if ( ! defined($fieldoptions{description}) || ! defined($options{description}) || $options{description} <= 1 ) {
			my $vtype=ref($value);
			if ( $vtype eq 'ARRAY' ) {
				$text .= wrapLine($value,$maxlength,abfprefix => $inckeyprefix);
			} elsif ( $vtype eq 'HASH' ) {
				my @items;
				for my $key (sort keys %{$value}) {
					push(@items,$key . " - " . $value->{$key});
				}
				$text .= wrapLine(\@items,$1,abfprefix => $inckeyprefix);
			} elsif ( $vtype eq '' ) {
				$text .= wrapLine([split /\s+/,$value],$maxlength,abfprefix => $inckeyprefix);
			}
		}

		if ( defined($options{description}) && $options{description} && defined($fieldoptions{description}) ) {
			if ( $options{description} == 1 ) {
				$text .= $inckeyprefix . "D: ";
			}
			$text .= $fieldoptions{description} . "\n";
		}
	}
	return $text;
}

=item $rext=outputTable(\@lines);

Renders a table with aligned columns. I<@lines> is an array of array refs
of the cell contents.

=cut

sub outputTable {
	my ($table,%options)=@_;
	if ( @{$table} == 0 ) {
		return '';
	}
	my $line1=$table->[0];
	my @max;
	for(my $n=0;$n<scalar(@{$line1});$n++) {
		$max[$n]=length($line1->[$n]);
	}
	for my $line (@{$table}) {
		for(my $n=0;$n<scalar(@max);$n++) {
			if ( length($line->[$n]) > $max[$n] ) {
				$max[$n]=length($line->[$n]);
			}
		}
	}
	my $text;
	for my $line (@{$table}) {
		my @line;
		for(my $n=0;$n<scalar(@max);$n++) {
			push(@line,align($line->[$n],$max[$n]));
		}
		$text .= join(' ',@line) . "\n";
	}
	return $text;
}

=item $length=stringLength($string,%options)

Returns the I<$length> of I<$string>, taking tabs into account.
Possible I<%options> are:

=over 4

=item B<tablength>

Set tablength (default: B<3>).

=back

=cut

sub stringLength {
	my ($s,%options)=@_;
	my $tablength=3;
	if ( $options{tablength} ) {
		$tablength=$options{tablength};
	}
	my $tab_=' ' x $tablength;
	$s =~ s/\t/$tab_/g;
	return length($s);
}

=item \@result=compactArray(\@array)

Compacts an I<@array> which means an array like

 I<@array> ('a','a','b','c','c','c')
 
is transformed to

 I<@result> ('2xa','b','3xc')

=cut

sub compactArray {
	my $old=shift;
	my @new;
	my $last;my $count=0;
	for my $i (@{$old}) {
		if ( $count ) {
			if ( $last eq $i ) {
				$count++;
			} else {
				if ( $count > 1 ) {
					push(@new,"${count}x$last");
				} else {
					push(@new,$last);
				}
				$last=$i;$count=1;
			}
		} else {
			$last=$i;$count=1;
		}
	}
	if ( $count ) {
		if ( $count > 1 ) {
			push(@new,"${count}x$last");
		} else {
			push(@new,$last);
		}
	}
	return \@new;
}


sub Dump {
	$Data::Dumper::Indent=1;
	$Data::Dumper::Sortkeys=1;
	return Dumper(@_);
}


=item $result=handleDbField(...)

=cut

sub handleDbField {
	my ($updates,$dbhash,$fieldname,$newvalue,$quotemethod,$text,$objectname)=@_;
	if ( $dbhash && %{$dbhash} ) {
		my $oldvalue=$dbhash->{$fieldname};
		if ( ! defined($oldvalue) and ! defined($newvalue) ) {
			return 1;
		}
		if ( $text ) {
			if (
				( defined($oldvalue) and ! defined($newvalue) ) ||
				( defined($newvalue) and ! defined($oldvalue) ) ||
				$oldvalue ne $newvalue
			) {
				out('d',"Updating field '$fieldname' on '$objectname'");
				push(@{$updates},"`${fieldname}`=" . &{$quotemethod}($newvalue));
				if ( defined($dbhash->{$fieldname}) && defined($oldvalue) ) {
					my $dblength=length($dbhash->{$fieldname});
					if ( defined($newvalue) && length($oldvalue) > 0 && substr($newvalue,0,$dblength) eq $dbhash->{$fieldname} ) {
						out('w',"Substring anomaly while updating field '$fieldname' on '$objectname' (length='$dblength'). Schema change required?");
					}
					out('d',"Old: $dbhash->{$fieldname}");
				}
				if ( defined($newvalue) ) {
					out('d',"New: $newvalue");
				}
			}
		} else {
			if (
				( defined($oldvalue) and ! defined($newvalue) ) ||
				( defined($newvalue) and ! defined($oldvalue) ) ||
				$oldvalue != $newvalue
			) {
				out('v',"Updating field '$fieldname' on '$objectname'");
				if ( defined($dbhash->{$fieldname}) ) {
					out('d',"Old: $dbhash->{$fieldname}");
				}
				if ( defined($newvalue) ) {
					out('d',"New: $newvalue");
				}
				push(@{$updates},"`${fieldname}`=" . &{$quotemethod}($newvalue));
			}
		}
	} else {
		if ( $text ) {
			push(@{$updates},&{$quotemethod}($newvalue));
		} else {
			if ( defined($newvalue) ) {
				push(@{$updates},$newvalue);
			} else {
				push(@{$updates},'NULL');
			}
		}
	}
	return 1;
}



=item \%tbd=retireBackups(\%backups,\@retentiontimes,%options)

Calculates the backups that can be removed from I<%backups> while keeping
given I<@retentiontimes> satisfied. I<%backups> is a hash with keys being
epoch values. I<@retentiontimes> is an array - a number of backups to be
kept representing each supported time unit:
(I<$hourly>,I<$daily>,I<$weekly>,I<$monthly>). On successful execution,
I<%tbd> will return epoch values to be deleted as keys and the respective
values of I<%backups> as values.

Possible I<%options> are:

=over 4

=item B<now> => I<epoch>

This options defines the time to be considered as "now" instead
of actually now. This is used for ftesting.

=back

Retention strategy:

=over 4

=item The latest backup is always kept.

=item Every rentention period is handled/satisfied separately.

=item I<@retentiontimes>=B<(0,0,0,0)> means: Only keep the latest backup.

=item Retention periods are handled per full hour/calendar day/calendar week/calendar month.

=item Retention of B<1> day means, keep one backup of of this day and one of the day before. Same goes for weeks, months and hours.

=back


=cut


sub retireBackups {
	my ($backups,$retentiontimes,%options)=@_;
	if ( ! $have_libdate_calc ) {
		out('w',"libdate-calc-perl is not installed. Unable to calc backup removal.");
		return undef;;
	}
	my $now;
	if ( defined($options{now}) ) {
		$now=$options{now};
	} else {
		$now=time();
	}
	out('d',"Retiring backups. Retention times: " . join(',',@{$retentiontimes}));

	
	# Bookkeeping
	my %weektree; # diff in weeks -> \@backups
	my %monthtree; # diff in months -> \@backups

	my %hourtree; # (delta_in_hours(now()-backup),\@backups)
	my %daytree; # (delta_in_days(now()-backup),\@backups)
	
	if ( ref($retentiontimes) ne 'ARRAY' || scalar(@{$retentiontimes}) != 4 ) {
		out('e',"retireBackups(): Invalid retention time arrayref");
		return undef;
	}
	my ($hours,$days,$weeks,$months)=@{$retentiontimes};

	my @n=localtime($now);
	my $nowhour=timelocal(0,0,$n[2],$n[3],$n[4],$n[5]); # Epoch of the first second of this hour (or the one given by $options{now})
	# Start of this week (0=monday, 6=sunday
	my $nwday=$n[6]-1;
	if ( $nwday == -1 ) {
		$nwday=6;
	}
	my $nowweek=$nowhour-$n[2]*3600-$nwday*3600*24;

	my %keep; # keys=epochs to keep
	my $first=1;
	
	my @backups=sort keys %{$backups};
	# Always keep the first backup 
	if ( @backups ) {
		$keep{$backups[$#backups]}++;
	}
	
	# Scan through the backup' epochs, make indices of the time differences to "now"
	for my $time (@backups) {
		# Backups of 'now' and from the future are never removed since this is considered
		# a bug.
		if ( $time > $now) {
			$keep{$time}++;
			next;
		}
		my @t=localtime($time);
		my $thishour=timelocal(0,0,$t[2],$t[3],$t[4],$t[5]);
		my $deltahours=int(sprintf("%.0f",($nowhour - $thishour)/3600)); # Due to leap-seconds, this might not be exactly an integer
		push(@{$hourtree{$deltahours}},$time);

		my $deltadays=Delta_Days($t[5]+1900,$t[4]+1,$t[3], $n[5]+1900,$n[4]+1,$n[3]);
		push(@{$daytree{$deltadays}},$time);

		# Start of this week (0=monday, 6=sunday
		my $wday=$t[6]-1;
		if ( $wday == -1 ) {
			$wday=6;
		}
		my $thisweek=$thishour-$t[2]*3600-$wday*3600*24;
		my $deltaweeks=int(sprintf("%0.f",($nowweek - $thisweek)/(3600*24*7)));
		push(@{$weektree{$deltaweeks}},$time);
		
		my $deltamonths=$n[5]*12+$n[4] - ($t[5]*12+$t[4]);
		push(@{$monthtree{$deltamonths}},$time);
	}

	# Iterate over all periods
	my @periods=(
		[$hours,\%hourtree],
		[$days,\%daytree],
		[$weeks,\%weektree],
		[$months,\%monthtree],
	);
	for my $period (@periods) {
		for my $delta (sort keys %{$period->[1]}) {
			if ( $delta <= $period->[0] ) {
				my @kept=grep { $keep{$_} } @{$period->[1]->{$delta}};
				if ( ! @kept ) {
					$keep{$period->[1]->{$delta}->[0]}++;
				}
			}
		}
	}

	my %tbd;
	for my $epoch (grep { ! $keep{$_} } keys %{$backups}) {
		$tbd{$epoch}=$backups->{$epoch};
	}	
	return \%tbd;
}


=item $xc=getXmlXc($text,%options)

Returns the xpathcontext I<$xc> associated to an XML string I<$test>.

Possible I<%options> are:

=over 4

=item B<ns> = I<\%namespaces>

Defines a hash of namespace shortcuts (keys) associated with
namespace URLs (values).

=item B<cache> = I<\%cache>

Fill in some interesting objects in the hash I<%cache>. This is being
used to cache reading XML files via L<Fb::IO::readXmlFile()> .

=item B<huge> => B<1>

Sent 'huge' option to XML parser to enable deeply nested XML files to
be parsed.

=back

An XPath context is returned on success, B<undef> on failure.

=cut

sub getXmlXc {
	my ($text,%options)=@_;
	my $parser;
	eval("\$parser=XML::LibXML->new();");
	if ( $@ ) {
		out('e',"Unable to create XML parser: $@");
		return undef;
	}
	my %parser_options=(
		no_network => 1,
		load_ext_dtd => 0,
		validation => 0,
   );
	if ( $options{huge} ) {
		$parser_options{'huge'}=1;
	}
	$parser->set_options(%parser_options);
	
	my $doc=$parser->parse_string($text);
	my $xc;
	eval "\$xc=XML::LibXML::XPathContext->new(\$doc);";
	if ( $@ ) {
		if ( ! $options{test} ) {
			out('e',"Unable to create XPath context");
		}
		return undef;
	}
	if ( $options{ns} ) {
		for my $name (sort keys %{$options{ns}}) {
			$xc->registerNs($name,$options{ns}->{$name});
		}
	}
	if ( defined($options{cache}) ) {
		$options{cache}->{root}=$doc->getDocumentElement;
		$options{cache}->{doc}=$doc;
		$options{cache}->{xc}=$xc;
		$options{cache}->{infotime}=time();
	}
	return $xc;
}


=back

=head2 Progress display

=over 4

=item $result=updateEtaStatus(\%progressrec,$command,@values)

Updates the progress information in I<%progressrec> and shows
progress of a multi-file copy operation. Possible I<$command>/<@values>
combinations are:

=over 4

=item B<init> I<files> I<bytes>

Initializes a I<$progressrec> for a (possible multi-file-) copy process.
The number of I<files> and their size in I<bytes> have to be given as
arguments.

=item B<initfile> I<filesize>

Registed the start of a copy process of a single file with a given I<filesize>.

=item B<updatefile> I<bytes>

Register progress during copying a single file. I<bytes> is the number of bytes
of the file that were copied since the last call to I<updateEtaStatus()>.

=item B<finishfile> {I<bytes>}

Mark the current file as completly copied. A number of I<bytes> might optionally
be given to indicate that not the whole file was copied but in fact only
that many I<bytes>. This will also correct the expected number of bytes of the
whole copy process.

=back

The whole copy process doesn't need to be finished. I<showEtaStatus(\%progressrec)>
can be called any time after the B<init>-call was made.

I<$result> == B<1>, only if no fatal error occured.

=cut

sub updateEtaStatus {
	my ($progressrec,$command,@values)=@_;
	if ( ! $progressrec ) {
		return 1;
	}
	if ( $command eq 'init' ) {
		%{$progressrec}=(
			files => $values[0], # Number of files to be copied
			bytes => $values[1], # Sum of all files' sizes
			filebytes => undef, # Size of the current file
			copiedfiles => 0, # Already copied files,
			copiedbytes => 0, # Already copied bytes,
			filecopiedbytes => undef, # Already copied bytes of the current file,
			start => time(), # Start of the whole copy process
			filestart => undef, # Start of the current file's copy process
		);
	} elsif ( $command eq 'initfile' ) {
		$progressrec->{filebytes}=$values[0];
		$progressrec->{filestart}=time();
		$progressrec->{filecopiedbytes}=0;
		$progressrec->{stephint}=[time(),0];
	} elsif ( $command eq 'updatefile' ) {
		$progressrec->{filecopiedbytes} += $values[0];
		showEtaStatus($progressrec);
		# Stephint contains 2 values:
		#  [0]: Time of the last 'updatefile' call
		#  [1]: Copied bytes of the current file during that time
		$progressrec->{stephint}=[time(),$progressrec->{filecopiedbytes}];
	} elsif ( $command eq 'finishfile' ) {
		if ( @values ) {
			my $diff=$progressrec->{filebytes}-$values[0];
			$progressrec->{bytes} -= $diff;
			$progressrec->{copiedbytes} += $values[0];
		} else {
			$progressrec->{copiedbytes} += $progressrec->{filebytes};
		}
		$progressrec->{filebytes}=undef;
		$progressrec->{filecopiedbytes}=undef;
		$progressrec->{filestart}=undef;
		$progressrec->{copiedfiles}++;
		delete($progressrec->{stephint});
	} else {
		out('e',"updateEtaStatus(): Invalid command '$command'");
		return 0;
	}
	return 1;
}


=item showEtaStatus(\%progressrec);

Show human readable status about the progress of a copy operation.

=cut

sub showEtaStatus {
	my ($p)=@_;
	# A full global statistic means that we know the number of files and the sum of their
	# sizes.
	my $fullglobalstat=0;
	if ( defined($p->{files}) && defined($p->{bytes}) ) {
		$fullglobalstat=1;
	}
	
	my $start;my $copied;my $bytes;
	my $add='';my $cacheetahr;
	if ( $fullglobalstat ) {
		$start=$p->{start};
		$copied=$p->{copiedbytes};
		if ( defined($p->{filecopiedbytes}) ) {
			# Add the per-file progress if possible
			$copied += $p->{filecopiedbytes};
		}
		$bytes=$p->{bytes};
	} else {
		$start=$p->{filestart};
		$copied=$p->{filecopiedbytes};
		$bytes=$p->{filebytes};
		$add='[f]';
	}
	my $byteshr=human_readable($bytes,'memory',space => '');
	my $copiedhr = human_readable($copied,'memory',space => '');
	my $tobecopied=$bytes-$copied;
	
	my $now=time();
	my $tdiff=$now-$start;

	# Just avoid div-by-0 when we just started
	if ( $tdiff == 0 || $copied == 0 ) {
		out('i',"Progress: copied=0, eta=?/?");
		return 1;
	}
	
	my $rate=$copied/$tdiff;
	my $ratehr = human_readable($rate,'memory',space => '',limit => 1000);
	
	my $etahr;my $endhr;
	if ( $copied ) {
		my $eta=int($tobecopied/$rate);
		$etahr = human_readable($eta,'timeinterval');
		$endhr=timeStamp('YYYYMMDD-hhmm',$now + $eta);
	} else {
		$etahr='?';
		$endhr='?';
	}

	my $stepetahradd='';my $stepratehradd='';
	if ( $p->{stephint} ) {
		# We have additional information provided by in-progress-measurements
		my $tdiffstep=$now-$p->{stephint}->[0];
		my $copiedstep=$p->{filecopiedbytes}-$p->{stephint}->[1];
		if ( $tdiffstep != 0 && $copiedstep != 0 ) {
			# We're not at time 0. Otherwise we could not devide by it
			my $steprate=$copiedstep/$tdiffstep;
			$stepratehradd="(" . human_readable($steprate,'memory',space => '',limit => 1000) . ")";
			my $stepeta=int($tobecopied/$steprate);
			$stepetahradd="(" . human_readable($stepeta,'timeinterval') . ")";
		}
	}

	my @info=();
	if ( defined($p->{files}) ) {
		push(@info,"f=$p->{copiedfiles}/$p->{files}");
	}
	push(@info,
		"s${add}=$copiedhr/$byteshr",
		"r=${ratehr}${stepratehradd}",
		"eta${add}=${etahr}${stepetahradd}",
	);
	out('i',"Progress: " . join(', ',@info));
	return 1;
}


=back

=cut

1;
