package Fb::IO;

use 5.006;
use strict;
use warnings;
use Fb::Data qw(hash_dir diffStrings bash_quote listEquality updateEtaStatus showEtaStatus getXmlXc max Dump);
use Fb::Out qw(out $VERBOSE $HIDDENDEBUG);
use File::Copy qw(copy);
use File::Temp qw(tempfile);
use File::stat;
use File::Find qw(find);
use Fcntl qw(S_IFMT S_IFREG S_IFBLK S_IFDIR LOCK_EX LOCK_UN LOCK_NB);

my $sendfile_available=0;
eval "use Sys::Syscall qw(sendfile);";
if ( $@ eq '' ) {
	$sendfile_available=1;
}
require Exporter;

our @ISA=qw(Exporter);
our @EXPORT_OK=qw(
	hash_mkdir
	readFile
	readXmlFile
	writeFile
	readDir
	iaexec
	$iaeexitcode
	prepareFtests
	readSums
	writeSums
	doFtest
	copyFile
	getFsObjectSize
	callFileLocked
	dryrunVerbose
);
our @EXPORT=qw(iaexec readFile writeFile);

my $fake_iaexec;
my $fake_iaexec_prefix=[];
our $iaeexitcode;

=over 4

=item $result=hash_mkdir($object,$path,$levels,%options);

Ensures (creates if necessary) a I<$levels>-level path relative to I<$path>
which is derived from the I<object>'s name.

I<$result> == 1, only if the directory exists after the call.

=cut

sub hash_mkdir {
	my ($object,$path,$levels,%options)=@_;
	my $dirs=hash_dir($object,$levels) || return 0;
	my $currentpath=$path;
	for my $dir (@{$dirs}) {
		$currentpath .= "/$dir";
		if ( ! -d $currentpath ) {
			if ( ! $options{'dryrun'} ) {
				if ( ! mkdir($currentpath) ) {
					Fb::Out::out('d',"Unable to mkdir($currentpath): $!");
					return 0;
				}
			} else {
				Fb::Out::out('i',"Would've created hash directory '$currentpath' now");
			}
		}
	}
	return 1;
}

=back

=head2 Simple IO

=over 4

=item $content=readFile($filename,%options)

Returns the I<$content> of the file I<$filename> or B<undef> in case
of an error. The I<$content> is a string containing the whole unmodified file.
Possible I<%options> are:

=over 4

=item B<encoding> (default: B<bytes>)

Set encoding for fileaccess. Possible values are:

=over 4

=item B<bytes>

=item B<utf8>

=back

=item B<format>

Possible formats are:

=over 4

=item B<string>

Just a string (default)

=item B<arrayref>

An array reference containing the file's lines with line terminators removed.

=item B<hashref>

A hash reference containing the file's lines with line terminators removed as keys of a hash with values being B<1>.

=back

=item B<dotexpand> => B<1>

If set, any occurence of B<${DOT}> is replaced by either a B<"."> character if
I<$ENV{'DOT'}> is set or B<""> if it's not.

=back

=cut

sub readFile {
	my ($filename,%options)=@_;
	my $format='string';
	if ( defined($options{format}) ) {
		$format=$options{format};
	}
	if ( $options{dotexpand} ) {
		my $rep='';
		if ( defined($ENV{DOT}) && $ENV{DOT} eq '.' ) {
			$rep='.';
		}
		$filename =~ s/\$\{DOT\}/$rep/g;
	}
	my $fh;
	my %openmap=(
		'utf8' => ':encoding(UTF-8)',
		'bytes' => '',
		'iso-8859' => ':encoding(iso-8859)',
	);
	my $encadd='';
	if ( defined($options{encoding}) ) {
		if ( defined($openmap{$options{encoding}}) ) {
			$encadd=$openmap{$options{encoding}};
		} else {
			out('e',"readFile(): Invalid encoding given '$options{encoding}'");
			return undef;
		}
	}
	out('dd',"Opening '$filename' with encoding '$encadd'");
	if ( ! open($fh,"<$encadd" ,$filename) ) {
		if ( ! $options{test} ) {
			Fb::Out::out('e',"Error reading file '$filename': $!");
		}
		return undef;
	}
	my $encoding='bytes';
	if ( $options{encoding} ) {
		$encoding=$options{encoding};
	}
	return _fh2content($fh,$format);
}

=item $result=_fh2content($fh,$format)

_fh2content is a generic function to read data from an open filehandle I<$fh>. $result is
depends on the given format I<$format>:

=over 4

=item B<string> Returns the the output of I<$fh> as a string without modifications.

=item B<arrayref> Returns the output as a array of lines with line terminators removed.

=item B<hashref> returns the output as a hashref with keys being the line-terminator-free lines of the output.

=item I<\$string> Puts the output unmodified into the string and returns it in I<$result>.

=item I<\@array> Puts the output as a list of lines with line terminators removed and retuns the array ref as I<$result>.

=item I<\%hash> Puts the output as one key per line with value B<1> into the hash I<%hash> and returns the hashref as I<$result>.

=item I<fh> Returns a filehandle to read the file content from.

=back

=cut

sub _fh2content {
	my ($fh,$format)=@_;
	my $content;
	my $direct; # Whether to return a string directory or just the reference
	if ( ref($format) eq 'SCALAR' ) {
		$content=$format;
		$$content='';
		$direct=0;
		$format='string';
	} elsif ( ref($format) eq 'ARRAY' ) {
		$content=$format;
		@{$content}=();
		$format='arrayref';
	} elsif ( ref($format) eq 'HASH' ) {
		$content=$format;
		$format='hashref';
	} elsif ( $format eq 'string' ) {
		my $temp='';
		$direct=1;
		$content=\$temp;
	} elsif ( $format eq 'arrayref' ) {
		my @temp=();
		$content=\@temp;
	} elsif ( $format eq 'hashref' ) {
		my %temp=();
		$content=\%temp;
	} elsif ( $format eq 'fh' ) {
		$?=0;
		return $fh;
	} else {
		out('e',"_fh2content(): Invalid format '$format'");
		#####  This should better be undef, right? iaexec Must be checked for this...
		return 0;
	}
	if ( $format eq 'string' ) {
		my $buf;
		while ( read($fh,$buf,65536) > 0 ) {
			$$content .= $buf;
		}
		close($fh);
		if ( $direct ) {
			return $$content;
		} else {
			return $content;
		}
	} elsif ( $format eq 'arrayref' ) {
		while ( my $l=<$fh> ) {
			chomp($l);
			push(@{$content},$l);
		}
		close($fh);
		return $content;
	} elsif ( $format eq 'hashref' ) {
		while ( my $l=<$fh> ) {
			chomp($l);
			$content->{$l}=1;
		}
		close($fh);
		return $content;
	}
}

=item $result=writeFile($filename,$content,%options)

Writes given I<$content> into file I<$filename>. Possible I<%options> are:

=over 4

=item B<dryrun>

Only simulate things.

=item B<encoding> (default: B<bytes>)

Set encoding for fileaccess (see perldoc binmode).

=item B<oldcontent>=I<string>

Old file's content. This will be compared to I<$content> to determine, if an update
is necessary.

=item B<testold>

If B<oldcontent> option is not given, this option being true will make writeFile()
read a possibly existing file first to determine, if writing is really necessary.
It's the same as writeFile($file,$content,oldcontent => readFile($file)).

=item B<writemode> (defaults to B<atomic>)

Possible values are:

=over 4

=item B<atomic>: Write new file, rename it to the correct filename after successful writing.

=item B<simple>: Just write the new file, remove if on error.

=item B<backup>: Copy the old file, write the new one, remove the backup on success.

=back

=item B<mode>

Sets the permission bits.

=item B<rcs>

A log message to be used when checking the old file into RCS. The mere presence of
this option triggers checking in the old file into RCS.

=back

Result returns one of these integers:

=over 4

=item B<0>: An error occured

=item B<1>: The file is in the requested state now - but nothing was actually written.

This is returned when the file content is actually the old content.

=item B<3>: The file was actually changed.

=back

=cut

sub writeFile {
	my ($filename,$content,%options)=@_;
	my $writemode='atomic';
	if ( $options{writemode} ) {
		$writemode=$options{writemode};
	}
	my $fh;
	if ( ref($content) eq 'ARRAY' ) {
		$content=join("\n",@{$content});
		if ( length($content) ) {
			$content .= "\n";
		}
	}
	my $encoding='bytes';
	if ( $options{encoding} ) {
		$encoding=$options{encoding};
	}
	my $oldcontent;
	if ( exists($options{oldcontent}) ) {
		$oldcontent=$options{oldcontent};
	} elsif ( $options{testold} || $options{dryrun} ) {
		$oldcontent=readFile($filename,encoding => $encoding,test => 1);
	}
	if ( defined($oldcontent) ) {
		if ( ref($oldcontent) eq 'ARRAY' ) {
			$oldcontent=join("\n",@{$options{oldcontent}});
			if ( length($oldcontent) ) {
				$oldcontent .= "\n";
			}
		}
		if ( $oldcontent eq $content ) {
			Fb::Out::out('d',"No changes, not writing '$filename'");
			return 1;
		}
		
	}
	if ( defined($oldcontent) ) {
		if ( $Fb::Out::VERBOSE > 2 ) {
			Fb::Out::out('dd',"File '$filename' 's changes:");
			diffStrings($oldcontent,$content);
		}
	} else {
		if ( $Fb::Out::VERBOSE > 2 ) {
			Fb::Out::out('dd',$content);
		}
	}
	if ( $options{dryrun} ) {
		Fb::Out::out('i',"Would've written file '$filename' now");
		return 1;
	}
	if ( $options{rcs} ) {
		my $fileq=quotemeta($filename);
		if ( -e $filename ) {
			my @initial=();
			if ( ! -e "$filename,v" ) {
				# Avoid ci asking for file description
				@initial=('-t-');
			}
			my $rcs_ok=iaexec(
				["/usr/bin/ci",'-l',@initial,"-m$options{rcs}",$filename],
				'exec',
			);
			if ( ! $rcs_ok ) {
				Fb::Out::out('e',"Checking '$filename' into RCS was requested but was unsuccessful");
				return 0;
			}
		} else {
			Fb::Out::out('d',"Ommiting checking non-existing file into RCS");
		}
		Fb::Out::out('d',"Successfully checked '$filename' into RCS");
	}
	my $newfilename;my $backupfilename;
	if ( $writemode eq 'atomic' ) {
		$newfilename="$filename.new.$$";
		if ( ! open($fh,"> $newfilename") ) {
			Fb::Out::out('e',"Unable to write file '$newfilename': $!");
			return 0;
		}
	} elsif ( $writemode eq 'backup' ) {
		$backupfilename="$filename.bak.$$";
		if ( ! copy($filename,$backupfilename) ) {
			Fb::Out::out('e',"Unable to create backup of file '$filename'");
			return 0;
		}
		if ( ! open($fh,"> $filename") ) {
			Fb::Out::out('e',"Unable to write file '$filename': $!");
			if ( ! rename($backupfilename,$filename) ) {
				Fb::Out::out('e',"Unable to restore backup file '$backupfilename' to '$filename': $!");
			}
			return 0;
		}
	} elsif ( $writemode eq 'simple' ) {
		if ( ! open($fh,"> $filename") ) {
			Fb::Out::out('e',"Unable to write file '$filename': $!");
			return 0;
		}
	} else {
		out('e',"writeFile(): Invalid write mode '$writemode'");
		return 0;
	}
	binmode($fh,":$encoding");
	print $fh $content;
	if ( ! close($fh) ) {
		Fb::Out::out('e',"Error closing file: $!");
		return 0;
	}
	if ( $writemode eq 'atomic' ) {
		if ( ! rename($newfilename,$filename) ) {
			Fb::Out::out('e',"Error moving file '$newfilename' into position");
			unlink($newfilename);
			return 0;
		}
	} elsif ( $writemode eq 'backup' ) {
		unlink($backupfilename);
	}
	if ( defined($options{mode}) ) {
		if ( ! chmod($options{mode},$filename) ) {
			out('w',"Unable to set permissions of '$filename'");
			return 0;
		}
	}
	return 3;
}


=item \%dentries=readDir($dir,%options)

Reads the dentries in directory I<$dir> as keys into hash(ref) I<%dentries>
with the exception of B<.> and B<..> .  Possible I<%options> are:

=over 4

=item B<filter> = I<regular_expression>

Defines an RE which has to match for a dentry to be included in the hash.

=item B<format>

Defines the output format of the dentry list. Possible Values are
B<arrayref> for a array reference and B<hashref> (default) for a
hash reference.

=item B<dotexpand> => B<1>

If set, any occurence of B<${DOT}> is replaced by either a B<"."> character if
I<$ENV{'DOT'}> is set or B<""> if it's not.

=back

=cut

sub readDir {
	my ($dir,%options)=@_;
	if ( $options{dotexpand} ) {
		my $rep='';
		if ( defined($ENV{DOT}) && $ENV{DOT} eq '.' ) {
			$rep='.';
		}
		$dir =~ s/\$\{DOT\}/$rep/g;
	}
	my $fh;
	if ( ! opendir($fh,$dir) ) {
		out('e',"Unable to opendir('$dir')");
		return undef;
	}
	my @dentries;
	if ( $options{filter} ) {
		my $filter= qr($options{filter});
		@dentries=grep { $_ =~ $filter } readdir($fh);
	} else {
		@dentries=grep { $_ !~ /^\.\.?$/ } readdir($fh);
	}
	closedir($fh);

	my $format='hashref';
	if ( defined($options{format}) ) {
		$format=$options{format};
	}
	if ( $format eq 'arrayref' ) {
		return [sort @dentries];
	} elsif ( $format eq 'hashref' ) {
		my %dentries;
		for my $dentry (@dentries) { $dentries{$dentry}=1; }
		return \%dentries;
	} else {
		out('e',"readDir(): invalid format '$format'");
		return undef;
	}
}


=item $xc=readXmlFile($filename,%options)

Returns the xpathcontext I<$xc> associated to an XML file I<$filename>.

Possible I<%options> are:

=over 4

=item B<ns> = I<\%namespaces>

Defines a hash of namespace shortcuts (keys) associated with
namespace URLs (values).

=item B<cache> = I<\$cachehash>

A reference to an a hash can be given. readXmlFile() will use
this hash to cache further requests to the same file.

=item B<test> => B<1>

Set this to B<1> to not handle XML file read failure as a problem.

=item B<dotexpand> => B<1>

Expands The string "B<${DOT}>" in the filename to either an empty string or
B<.> - depending on the content of I<$ENV{'DOT'}>.
dotexpand is done in this method because of the stat()-backed
caching for which the filename needs to be known before readFile()ing.

=item B<huge> => B<1>,

Enable parsing deeply nested XML files.

=back

An XPath context is returned on success, B<undef> on failure.

=cut

sub readXmlFile {
	my ($filename,%options)=@_;

	if ( $options{dotexpand} ) {
		my $rep='';
		if ( defined($ENV{DOT}) && $ENV{DOT} eq '.' ) {
			$rep='.';
		}
		$filename =~ s/\$\{DOT\}/$rep/g;
	}

	my @stat=CORE::stat($filename);
	if ( ! @stat ) {
		if ( ! $options{test} ) {
			out('e',"readXmlFile(): Unable to access file '$filename'");
		}
		return undef;
	}
	if ( defined($options{cache}) ) {
		if ( $options{cache}->{mtime} ) {
			if ( $options{cache}->{mtime} eq $stat[9] ) {
				return $options{cache}->{xc};
			}
		}
	}

	my $text=readFile($filename,format => 'string',test => $options{test}) || return undef;
	my $result=getXmlXc($text,cache => $options{cache},ns => $options{ns},test => $options{test},huge => $options{huge});
	if ( $result && defined($options{cache}) ) {
		$options{cache}->{mtime}=$stat[9];
	}
	return $result;
}



=back

=head2 Executing commands

=over 4

=item $result=iaexec($cmdline,$mode,%options)

Executes command I<$cmdline>. I<$cmdline> may be given as an
arrayref to avoid executing a shell first. In any case, the exitcode of a
executed command is returned via the I<$iaeexitcode> global variable which means
"The whole exit code" as returned by wait() incl. a potentially received signal.
Several options may be given:

=over 4

=item B<mode>

Defines the behaviour of iaexec. Possible values are:

=over 4

=item B<test>

Just execute the command, return B<1> on exitcode == 0 and B<0> on everything else.
This mode is used to determin things - a failing command is not considered an error
but just a result.

=item B<exec> 

Same as B<test> but the command is expected to return successfully. This is why defaults
for B<stdout> and B<stderr> are slightly different.

=item B<writepipe>

Executes the command as a pipe to write to. I<$result> is an open filehandle.
in this case. This I<$mode>'s behaviour is influenced by the existence of the
B<input>-option. 

=item B<readpipe> 

Executes the command as a pipe to read from. I<$result> depends on the value of the B<format> options.

=back

=item B<stdout>

Possible values are:

=over 4

=item B<hide>

Never show anything the command outputs via STDOUT. this will not affect
data returned in B<readpipe> mode.

=item B<show>

Show output of the command on STDOUT. The output is a dublicate of
the data returned when B<readpipe> mode.

=item B<try>

STDOUT is stored until the command terminates. On failure the output is shown
on STDOUT. It is shown even if mode is B<readpipe> then.

=item B<redirect:>I<filename>

STDOUT is written to file I<filename>.

=back

Default is B<hide> if I<$mode> == B<test>, B<try> if I<$mode> == B<readpipe> or B<exec> and B<show> if I<$mode> == B<writepipe>.

=item B<stderr>

Possible values are:

=over 4

=item B<hide>

Hide and ignore everything via STDERR from the command.

=item B<show>

Show output of STDERR unfiltered on the terminal. This way it cannot be returned in
B<readpipe> mode.

=item B<include>

Redirect STDERR and handle it like STDOUT.

=back

Default is B<include> if I<$mode> == B<exec> and B<show> otherwise.

=item B<format>

Defines the format of data being returned in case of I<$mode> == B<readpipe>. Possible formats are:

=over 4

=item B<fh>

Returns just the filehandle of the pipe.

=item B<arrayref>

A reference to an array of lines (without line termination).

=item B<hashref>

A reference to a hash of lines (without line termination).

=item B<string>

A string containing the whole output.

=item I<\$string>, I<\@array>

Output is put in the format given by the reference into the referred object. Return code will be as in I<$mode> == 'test'.

=item I<\%hash>

Output is put into the referred I<%hash> *without* clearing the hash first.

=back

=item B<input> => B<\@array>

An I<@array> of lines to be sent to a B<writepipe>. When given, iaexec()
returns the success of the command (true, if the exitcode was B<0>, false
otherwise).

=item B<description>

A textual description of the action, this command should accomplish. It will
be used to print an error message in case of failure.  For the error message
to make sense, the format should be like this:

 Creating PTDB group
 
not

 PTDB group is being created

=item B<failure> = I<\%hash>|I<\@array>|I<\&code>

This option is used to define, which result codes are considered failures. It can be
either a hashref with keys being integer values, an arrayref with integer values or a code reference
which is called with the exitcode ($exitvalue >> 8). The code returning B<1> is considered a
failure, B<0> is success.

This value is used by the _iaexec_try() method which is called in variety of places:

=over 4

=item *

To decide, when a stderr/stdout-try is satisfied and the stored output should be displayed. 

=item *

To decide, if the result should be undef instead of the stored readpipe output.

=item *

To decide the return value in test mode.


=back

=item B<dryrun> => B<1>

Simulate executing the command successfully (works only in mode 'exec' and 'test', aborts with
an error in other modes). The command line in shown in that case.

=back

=cut

sub iaexec {
	my ($command,$mode,%options)=@_;
	$iaeexitcode=0;
	if ( ! defined($mode) ) {
		Fb::Out::out('e',"iaexec(): Missing mode key");
		return undef;
	}
	my $format;
	if ( $mode eq 'readpipe' ) {
		$format=$options{format};
		if ( ! defined($format) ) {
			Fb::Out::out('e',"iaexec(): No format given in 'readpipe' mode.");
			return undef;
		}
	} else {
		my $buffer='';
		$format=\$buffer;
	}

	# Do stuff for verbose user information
	my $command_=$command;
	if ( ref($command) eq 'ARRAY' ) {
		$command_=join(' ',map { bash_quote($_) } @{$command});
	}
	if ( $options{'description'} ) {
		Fb::Out::out('v',$options{'description'});
	}
	if ( $Fb::Out::VERBOSE > 2 ) {
		Fb::Out::out('dd',"Executing: " . Dump($command));
	} else {
		Fb::Out::out('d',"Executing $command_");
	}


	# Determine stderr/stdout-output policy
	my $stdout;my $stderr;
	if ( $mode eq 'test' ) {
		$stdout='try';$stderr='hide';
	} elsif ( $mode eq 'exec' ) {
		$stdout='try';$stderr='include';
	} elsif ( $mode eq 'readpipe' ) {
		$stdout='try';$stderr='show';
	} elsif ( $mode eq 'writepipe' ) {
		$stdout='show';$stderr='show';
	} else {	
		Fb::Out::out('e',"iaexec(): Invalid mode '$mode'");
		return undef;
	}
	if ( $options{'stdout'} ) { $stdout=$options{'stdout'}; }
	if ( $options{'stderr'} ) { $stderr=$options{'stderr'}; }
	# Data block containing all necessary Information for exec'ing
	my $execi={
		command => $command,
		command_ => $command_,
		stderr => $stderr,
		stdout => $stdout,
		stdoutbackup => undef,
		stderrbackup => undef,
		pipeneeded => 1,
		errors => [],
		mode => $mode,
		fh => undef,
		input => $options{input}, # This will contain writepipe data to be sent to the run program - as arrayref
	};
	if ( defined($options{input}) && ref($options{input}) eq '' ) {
		$execi->{input}=[split /\n/,$options{input}];
	}

	my $opensuccess=_iaexec_redirect_enable($execi);
	if ( $opensuccess ) {
		$opensuccess &= _iaexec_open($execi);
	}
	_iaexec_redirect_disable($execi);
	if ( ! $opensuccess ) {
		$iaeexitcode=65536;
		return undef;
	}
	
	if ( $execi->{mode} eq 'writepipe' ) {
		my $fh=$execi->{fh};
		if ( $options{input} ) {
			for my $l (@{$execi->{input}}) {
				print $fh "$l\n";
			}
			close($fh);
			$iaeexitcode=$?;
			return _iaexec_try($iaeexitcode,\%options);
		} else {
			return $fh;
		}
	}
	
	my $output;
	if ( $execi->{pipeneeded} ) { 
		# Read output from FD
		$output=_fh2content($execi->{fh},$format);
		if ( ! $fake_iaexec ) {
			# If fake_iaexec is used, the exitcode was already delivered
			$iaeexitcode=$?;
		}
	}
	my $tryresult=_iaexec_try($iaeexitcode,\%options);

	# dump output in case of stdout-try-policy
	if ( $execi->{stdout} eq 'try' ) {
		if ( ! $tryresult ) {
			# This is slighly incorrect if $format is (a) hashref
			_iaexec_output($output);
		}
	}
	
	if ( $execi->{mode} eq 'test' || $execi->{mode} eq 'exec' ) {
		return $tryresult;
	} elsif ( $execi->{mode} eq 'readpipe' ) {
		if ( ref($format) ) {
			# read data resides in referred variable, return interpreted result
			return $tryresult;
		} else {
			# Return data directly - or undef in case of failure
			if ( $tryresult ) {
				return $output;
			} else {
				return undef;
			}
		}
	}
}

=item _iaexec_redirect_enable($execi)

Helper method for starting stdout/stderr-redirections before actually
exec()ing commands in iaexec().

=cut

sub _iaexec_redirect_enable {
	my $execi=shift;
	my $success=1;
	# Redirect STDOUT if necessary
	if ( $execi->{mode} eq 'test' || $execi->{mode} eq 'exec' || $execi->{mode} eq 'writepipe' ) {
		if ( $execi->{stdout} eq 'hide' ) {
			open($execi->{stdoutbackup},'>&STDOUT');
			open(STDOUT,'>/dev/null');
			push(@{$execi->{errors}},['dd',"iaexec(): Hiding STDOUT"]);
			$execi->{pipeneeded}=0;
		} elsif ( $execi->{stdout} =~ /^redirect:(.*)$/ ) {
			my $outputfile=$1;
			open($execi->{stdoutbackup},'>&STDOUT');
			if ( open(STDOUT,"> $outputfile") ) {
				push(@{$execi->{errors}},['dd',"iaexec(): Redirecting STDOUT to '$outputfile'"]);
			} else {
				push(@{$execi->{errors}},['e',"iaexec(): Unable to redirect STDOUT to '$outputfile'"]);
				$success=0;
			}
			$execi->{pipeneeded}=0;
		} elsif ( $execi->{stdout} eq 'show' ) {
			$execi->{pipeneeded}=0;
		}
	}

	# Redirect STDERR, check stderr-option for validity
		if ( $execi->{stderr} eq 'include' ) {
			# has to be done in _iaexec_open
		} elsif ( $execi->{stderr} eq 'hide' ) {
			open($execi->{stderrbackup},'>&STDERR');
			open(STDERR,'>/dev/null');
			push(@{$execi->{errors}},['dd',"iaexec(): Hiding STDERR"]);
		} elsif ( $execi->{stderr} eq 'show' ) {
			# nothing to be done
		} elsif ( $execi->{stderr} =~ /^redirect:(.*)$/ ) {
			if ( ! $fake_iaexec ) {
				my $outputfile=$1;
				open($execi->{stderrbackup},'>&STDERR');
				if ( open(STDERR,"> $outputfile") ) {
					push(@{$execi->{errors}},['dd',"iaexec(): Redirecting STDERR to '$outputfile'"]);
				} else {
					push(@{$execi->{errors}},['e',"iaexec(): Unable to redirect STDERR to '$outputfile'"]);
					$success=0;
				}
			}
		} else {
			push(@{$execi->{errors}},['e',"iaexec(): Invalid stderr-option '$execi->{stderr}'"]);
			$success=0;
		}
	return $success;
}

=item _iaexec_redirect_disable($execi)

Helper method to reset redirections after iaexec()'s pipe-opening and
re-play out() calls.

=cut

sub _iaexec_redirect_disable {
	my $execi=shift;
	# Restore STDERR in case it was redirected
	if ( $execi->{stderrbackup} ) {
		push(@{$execi->{errors}},['dd',"iaexec(): Restoring STDERR"]);
		close(STDERR);
		open(STDERR,">&",$execi->{stderrbackup});
		close($execi->{stderrbackup});
		$execi->{stderrbackup}=undef;
	}
	# Restore STDOUT in case it was redirected
	if ( $execi->{stdoutbackup} ) {
		push(@{$execi->{errors}},['dd',"iaexec(): Restoring STDOUT"]);
		close(STDOUT);
		open(STDOUT,">&",$execi->{stdoutbackup});
		close($execi->{stdoutbackup});
		$execi->{stdoutbackup}=undef;
	}
	
	# play errors stored in possible STDOUT/STDERR-blackout
	for my $error (@{$execi->{errors}}) {
		Fb::Out::out(@{$error});
	}
	@{$execi->{errors}}=();
	return 1;
}

=item _iaexec_open($execi)

Helper method to open the read- or writepipe of the executed iaexec()
command.

=cut

sub _iaexec_open {
	my $execi=shift;
	if ( $Fb::Out::VERBOSE > 2 ) {
		push(@{$execi->{errors}},['d',"_iaexec_open(): " . Dump($execi)]);
	}
	
	# Make given command argument into an array, taking redirections into account
	my @command;
	if ( $execi->{stderr} eq 'include' ) {
		# Currently, we have to use bash to combine stderr and stdout
		if ( ref($execi->{command}) eq '' ) {
			push(@{$execi->{errors}},['dd',"iaexec(): Redirecting STDERR via shell to be included in readpipe"]);
			@command=("$execi->{command} 2>&1");
		} else {
			push(@{$execi->{errors}},['dd',"iaexec(): Converting vector-exec to commandline to redirect STDERR via shell"]);
			my $command=join(' ',map { bash_quote($_) } @{$execi->{command}}) . " 2>&1";
			@command=($command);
		}
	} else {
		if ( ref($execi->{command}) eq '' ) {
			# open() will treat this as a bash command as it is intended
			@command=($execi->{command});
		} elsif ( @{$execi->{command}} == 1 ) {
			# open() will treat this as a bash command, which we do not want but there's
			# currently no way around it
			@command=bash_quote($execi->{command}->[0]);
		} else {
			# open() with vector arguments - perfect
			@command=@{$execi->{command}};
		}
	}

	if ( $execi->{mode} eq 'readpipe' ||  $execi->{mode} eq 'exec' ||  $execi->{mode} eq 'test' ) {
		if ( ! $fake_iaexec ) {
			if ( $execi->{pipeneeded} ) {
				if ( ! open($execi->{fh},'-|',@command) ) {
					push(@{$execi->{errors}},['e',"Unable to execute $execi->{command_}"]);
					return 0;
				}
			} else {
				$iaeexitcode=system(@command);
			}
		} else {
			$iaeexitcode=&{$fake_iaexec}(@{$fake_iaexec_prefix},$execi->{command},'readpipefh',\$execi->{fh});
		}
	} elsif ( $execi->{mode} eq 'writepipe' ) {
		if ( ! $fake_iaexec ) {
			if ( ! open($execi->{fh},'|-',@command) ) {
				push(@{$execi->{errors}},['e',"Unable to (wp)-execute $execi->{command_}"]);
				return 0;
			}
		} else {
			if ( defined($execi->{input}) ) {
				$iaeexitcode=&{$fake_iaexec}(@{$fake_iaexec_prefix},$execi->{command},'writepipe',$execi->{input});
			} else {
				# We currently do not know how to do that correctly :-(
				open($execi->{fh},"> /dev/null");
				$iaeexitcode=0;
				push(@{$execi->{errors}},['w',"Unable to correctly fake $execi->{command_} in writepipe mode without given input."]);
			}
		}
	} else {
		return 0;
	}
	return 1;
}

=item _iaexec_output($content)

Helper method for printing data previously read from a pipe
in different data formats.

=cut

sub _iaexec_output {
	my $content=shift;
	if ( ref($content) eq '' ) {
		print $content;
	} elsif ( ref($content) eq 'SCALAR' ) {
		print $$content;
	} elsif ( ref($content) eq 'ARRAY' ) {
		print join("\n",@{$content});
		if ( @{$content} ) {
			print "\n";
		}
	} elsif ( ref($content) eq 'HASH' ) {
		print join("\n",sort keys %{$content});
		if ( scalar(keys %{$content}) ) {
			print "\n";
		}
	}
	return 1;
}

=item $result=_iaexec_try($exitcode,\%options)

Helper method to determine success of a command. Default is to return success (I<$result> == B<1>)
on I<$exitcode> == B<0> and failure (I<result> == B<0>) otherwise. This can be changed
via iaexec()'s I<%options> that are forwarded to this method.

=cut

sub _iaexec_try {
	my ($exitcode,$options)=@_;
	if ( ! $options->{failure} ){
		if ( $exitcode != 0 ) {
			return 0;
		} else {
			return 1;
		}
	} else {
		my $failure=$options->{failure};
		my $mode=ref($failure);
		my $exitvalue=$exitcode >> 8;
		if ( $mode eq '' ) {
			if ( $exitvalue == $failure ) {
				return 0;
			}
		} elsif ( $mode eq 'HASH' ) {
			if ( $failure->{$exitvalue} ) {
				return 0;
			}
		} elsif ( $mode eq 'ARRAY' ) {
			for my $v (@{$failure}) {
				if ( $v eq $exitvalue ) {
					return 0;
				}
			}
		} elsif ( $mode eq 'CODE' ) {
			if ( &{$failure}($exitvalue) ) {
				return 0;
			}
		} else {
			Fb::Out::out('e',"Invalid value of failure option in iaexec(), returning failure to get attention.");
			return 0;
		}
		return 1;
	}
}

=back

=head2 Checksum handling on permanent storage

=over 4

=item \%sums=readSums($filename)

=cut

sub readSums {
	my ($filename,%options)=@_;
	if ( ! -e $filename ) {
		return {};
	}
	my $sums=readFile($filename,format => 'arrayref') || return undef;
	out('d',"Read '$filename', got '" . scalar(@{$sums}) . "' lines");
	my %sums;
	my $errors=0;
	for my $l (@{$sums}) {
		if ( $l !~ /^([a-f0-9]+) \*?(.*)$/ ) {
			out('e',"Error reading sums file '$filename'");
			out('i',"Offending line: '$l'");
			$errors++;
			next;
		}
		$sums{$2}=$1;
	}
	if ( $errors ) {
		return undef;
	}
	return \%sums;
}

=item $result=writeSums($filename,\%sums,%options)

Writes a hashsum file if there are any checksums in the
I<%sums>-map. Possible I<%options> are:

=over 4

=item B<dryrun>

Only simmulate things

=item B<removeempty>

Set to B<1>, the sums file will be removed, if no I<%sums>
are available.

=back

=cut

sub writeSums {
	my ($filename,$sums,%options)=@_;
	my @sums;
	for my $file (sort keys %{$sums}) {
		push(@sums,"$sums->{$file} *$file");
	}
	if ( ! @sums && $options{removeempty} ) {
		if ( -e $filename ) {
			if ( $options{dryrun} ) {
				out('i',"Would've remove checksum file '$filename' now");
				return 1;
			} else {
				if ( unlink($filename) ) {
					return 1;
				} else {
					out('e',"Error removing checksum file '$filename'");
					return 0;
				}
			}
		}
	}
	return writeFile($filename,\@sums,testold => 1,dryrun => $options{dryrun});
}

=back

=head2 Cool IO stuff

=over 4

=item $result=doFtest($filename,%options)

Do functional test by running a command given in the first line of the file I<$filename>,
test the exitcode against the numerical value in line 2 and the combined STDERR and STDOUT output
against the rest of the I<$filename> content. Possible I<%options> are:

=over 4

=item B<update>

If set to B<1>, on wrong results, the user is asked to confirm that the current results are
correct. The content of I<$filename> is replaced then.

=item B<test> = B<1>

Do not show output. Just indicate test result via I<$result>.

=item B<commandmodifier> => &sub(\$commandline)

A sub to call for pre-processing the I<$commandline> before executing.

=item B<severity> => I<\$severity>

A ref to a scalar into which the severity of a problem is written upon failure.

=item B<defaultseverityoutput> => B<n>

Sets default severity to be returned upon failure (Otherwise: 1)

=item B<defaultseverityexit> => B<n>

Sets default severity to be returned upon wrong exit code (Otherwise: 2)

=item B<dryrun> => B<1>

Only simulate writing files on update.

=back

I<$result> == B<1> if the test passed or if the if didn't, B<update> option was given and the new
test file was successfully written.

=cut

sub doFtest {
	my ($filename,%options)=@_;
	REDO:
	my $severity_output=1;
	if ( defined($options{defaultseverity_output}) ) {
		$severity_output=$options{defaultseverity_output};
	}
	my $severity_exit=2;
	if ( defined($options{defaultseverity_exit}) ) {
		$severity_exit=$options{defaultseverity_exit};
	}
	my @old_file=();my @old_header=();
	readFile($filename,format => \@old_file) || return 0;
	# Split the header lines from the content
	my $foundstart=0;
	my @old_content=@old_file;
	while ( scalar(@old_content) && ! $foundstart ) {
		my $l=shift(@old_content);
		push(@old_header,$l);
		if ( $l eq '' ) {
			$foundstart=1;
		}
	}
	my @header=@old_header;
	my $commandline=shift(@header);
	my $old_exitcode;my $exitcodemissmatch=0;
	for my $l (@header) {
		if ( $l =~ /^exitcode=([0-9]+|\*)$/ ) {
			$old_exitcode=$1;
			if ( $old_exitcode eq '*' ) {
				$old_exitcode=undef;
			} else {
				out('d',"Asserting exit code '$old_exitcode' for test '$filename'");
			}
		} elsif ( $l =~ /^severity_output=([0-9]+)$/ ) {
			$severity_output=$1;
		} elsif ( $l =~ /^severity_exit=([0-9]+)$/ ) {
			$severity_exit=$1;
		} elsif ( $l eq '' ) {
			# This is the space after the header
		} else {
			out('e',"Invalid ftest header line in '$filename': '$l'");
			return 0;
		}
	}
	if ( ! defined($old_exitcode) ) {
		out('d',"Exit code for test '$filename' will be ignored.");
	}

	if ( defined($options{commandmodifier}) ) {
		&{$options{commandmodifier}}(\$commandline) || return 0;
	}

	my @new_content;
	out('v',"CMD: $commandline");
	iaexec($commandline,'readpipe',format => \@new_content,stderr => 'include',stdout => 'include');
	
	# Undefined exitcode means: it doesn't matter.
	my $new_exitcode=$iaeexitcode >> 8;
	if ( defined($old_exitcode) && $old_exitcode != $new_exitcode ) {
		if ( ! $options{test} ) {
			out('v',"Invalid exitcode ($old_exitcode != $new_exitcode) for '$filename' ");
			$exitcodemissmatch=1;
		}
	}
	my $diff;
	my $outputmissmatch=0;
	if ( ! diffStrings(\@old_content,\@new_content,getdiff => \$diff) ) {
		$outputmissmatch=1;
		if ( ! $options{test} ) {
			out('w',"Content differing for '$filename'");
			if ( $Fb::Out::VERBOSE > 1 || $options{update} ) {
				print $diff;
			}
		}
	}
	my $severity=0;
	if ( $outputmissmatch ) {
		$severity=max($severity,$severity_output);
	}
	if ( $exitcodemissmatch ) {
		$severity=max($severity,$severity_exit);
	}
	
	if ( $severity && $options{update} ) {
		if ( $exitcodemissmatch ) {
			out('e',"Update requested but exitcode was wrong (want:$old_exitcode,got:$new_exitcode). No automatic upgrade.");
			return 0;
		}
		print "Update or re-do test '$filename' ? (y|r|N) :";
		my $answer=<STDIN>;chomp($answer);
		if ( $answer =~ /^(r|redo)$/i ) {
			goto REDO;
		} elsif ( $answer =~ /^(y|yes)$/i ) {
			my @new_file=(@old_header,@new_content);
			if ( writeFile($filename,\@new_file,oldcontent => \@old_file,dryrun => $options{dryrun}) ) {
				out('i',"Update applied successfully");
				return 1;
			} else {
				return 0;
			}
		}
		out('i',"No update applied.");
		return 1;
	}
	if ( $severity && defined($options{severity}) ) {
		${$options{severity}}=$severity;
	}
	return ($severity == 0);
}

=item $result=copyFile($sourcefilename,$targetfilename,%options);

Copy file I<$sourcefilename> to I<$targetfilename>. I<$targetfilename> is
*not a foldername* . Focus is on building a checksum of the file in transit.
If a checksum is calculated and one was available in the source folder's
md5sums.txt, it's compared and a error is returned on failure.
Possible I<%options> are:

=over 4

=item B<progress> => I<\%progressrec>

Calculate and output progress during copy.

=item B<checksums> => I<\%checksums>

Update checksum hash I<%checksumhash> with a digest calculated during copying.
Key is the dentry name of the I<$targetfilename>

=item B<checksum> => I<\$checksum>

Update checksum string I<$checksum>.

=item B<checksumfile> => B<1>

Update the checksum file on the target location.

=item B<checksumtype> => (B<md5>|B<sha1>|B<sha256>|B<sha512>)

=item B<trustoldchecksum> => B<1>

Try to read old checksum from source file's folder and treat it
as if it was just calculated.

=back

The method will choose the fastest possible copy mode.

I<$result> == B<1>, only if no error occured.

=cut

sub copyFile {
	my ($source,$target,%options)=@_;

	my $pv_available=(-x "/usr/bin/pv");
	my $digesttype='md5';
	if ( defined($options{checksumtype}) ) {
		if ( $options{checksumtype} =~ /^(md5|sha1|sha256|sha512)$/ ) {
			$digesttype=$1;
		} else {
			out('e',"Unknown '$options{checksumtype}'");
			return 0;
		}
	}
	my $digestbinary="/usr/bin/${digesttype}sum";
	my $sourceq=bash_quote($source);
	my $targetq=bash_quote($target);
	my $starttime=time();
	my $mode;


	$source =~ /^(?:(.*)\/)?([^\/]+)$/;
	my $sourcefolder=$1;my $sourcedentry=$2;
	if ( ! defined($sourcefolder) ) {
		$sourcefolder=".";
	}
	$target =~ /^(?:(.*)\/)?([^\/]+)$/;
	my $targetfolder=$1;my $targetdentry=$2;
	if ( ! defined($targetfolder) ) {
		$targetfolder=".";
	}

	my $targetisblock=0;my $newornot='.new';
	my $sourcesize=getFsObjectSize($source);
	if ( -b $target ) {
		out('v',"Target file is block device, no atomic copy possible.");
		$targetisblock=1;
		my $targetsize=getFsObjectSize($target);
		if ( $targetsize < $sourcesize ) {
			out('e',"Target '$target' is a block device but smaller than the source '$source'");
			return 0;
		}
		$newornot='';
	}
	

	
	my $oldchecksum; # Old checksum as read from the respective source folder
	
	my $somesums=( defined($options{checksums}) || defined($options{checksum}) || $options{checksumfile} ); 
	if ( $somesums ) {
		my $oldsums=readSums("$sourcefolder/${digesttype}sums.txt");
		if ( defined($oldsums->{$sourcedentry}) ) {
			$oldchecksum=$oldsums->{$sourcedentry};
		}
	}
	
	# Choose a method depending (mostly) on, if we need to calculate the
	# file's checksum during transfer.
	if ( $somesums && ! ( $options{trustoldchecksum} && defined($oldchecksum) ) ) {
		if ( defined($options{progress}) ) {
			if ( $pv_available ) {
				$mode="pv-digest";
			} else {
				out('w',"copyFile(): pv not available. No progress");
				$mode="tee-digest";
			}
		} else {
			$mode="tee-digest";
		}
	} else {
		if ( $sendfile_available ) {
			# This is the most efficient way - esp. in the light of meltdown protection
			$mode="sendfile";
		} else {
			# We call an external binary which should speed things up
			out('w',"Sendfile not available, libsys-syscall-perl seems to be missing.");
			$mode="cp";
		}
	}

	out('i',"Copying to target file '$targetdentry'");
	if ( $options{dryrun} ) {
		out('i',"Would've copied '$source' to '$target' using mode '$mode' now");
		$target =~ /^(?:(.*)\/)?([^\/]+)$/;
		my $targetfolder=$1;my $targetdentry=$2;
		if ( defined($options{checksum}) ) {
			$options{checksum}='feedbabe' x 4;
		}
		if ( defined($options{checksums}) ) {
			$options{checksum}->{$targetdentry}='feedbabe' x 4;
		}
		if ( $options{checksumfile} ) {
			out('i',"Would've updated checksum file '$targetfolder/${digesttype}sums.txt");
		}
		out('e',"Unable to simulate copy process any further. Returning error");
		return 0;
	}

	# The checksum is alway written to a tempfile 
	my $tempfile;my $tempfileq;
	if ( $somesums && ! ( $options{trustoldchecksum} && defined($oldchecksum) ) ) {
		if ( ! -x $digestbinary ) {
			out('e',"Checksum requested but digest binary '$digestbinary' is not available");
			return 0;
		}
		my $tempfh;
		($tempfh,$tempfile)=tempfile('media_copy_XXXXX',DIR => '/tmp');
		if ( ! $tempfile ) {
			out('e',"Unable to create tempfile for in-flight md5summing during copy of '$source'");
			return 0;
		}
		close($tempfh);
		$tempfileq=quotemeta($tempfile);
	}

	my $result;
	out('d',"Copying '$source' to '$target$newornot' using method '$mode'");
	if ( $mode eq 'pv-digest' ) {
		$result=iaexec(
			"set -o pipefail;/usr/bin/pv $sourceq | /usr/bin/tee $targetq$newornot | $digestbinary -b > $tempfileq",
			'exec',
			stdout => 'show',
			stderr => 'show',
		);
		if ( $result ) {
			updateEtaStatus($options{progress},'initfile',$sourcesize);
			updateEtaStatus($options{progress},'finishfile');
			showEtaStatus($options{progress});
		}
	} elsif ( $mode eq 'pv' ) {
		$result=iaexec(
			"/usr/bin/pv $sourceq > /usr/bin/tee $targetq$newornot",
			'exec',
			stdout => 'show',
			stderr => 'show',
		);
		if ( $result ) {
			updateEtaStatus($options{progress},'initfile',$sourcesize);
			updateEtaStatus($options{progress},'finishfile');
			showEtaStatus($options{progress});
		}
	} elsif ( $mode eq 'tee-digest' ) {
		$result=iaexec(
			"set -o pipefail;/usr/bin/tee < $sourceq $targetq$newornot | $digestbinary -b > $tempfileq",
			'exec',
			stdout => 'show',
			stderr => 'show',
		);
		# This method is only selected, if no progress is wanted
	} elsif ( $mode eq 'sendfile' ) {
		# This should be the fastest way on linux
		my $in;my $out;
		if ( ! open($in,"< $source") ) {
			out('e',"Unable to open source file '$source' for reading: $!");
			return 0;
		}
			
		if ( ! open($out,"> $target$newornot") ) {
			out('e',"Unable to open target file '$target$newornot' for writing: $!");
			return 0;
		}
		
		my $stepsize=1024*1024*1024; # Transfer size per sendfile()-syscall

		# We use this specilized function instead of -s to include block devices.
		updateEtaStatus($options{progress},'initfile',$sourcesize);

		my $last_copied=0;
		do {
			$last_copied=sendfile(fileno($out),fileno($in),$stepsize);
			if ( $last_copied > 0 ) {
				updateEtaStatus($options{progress},'updatefile',$last_copied);
			}
		} until ( $last_copied < 1 );
		updateEtaStatus($options{progress},'finishfile');
		if ( $last_copied == -1 ) {
			out('w',"sendfile() failed ($!). Disabling. Retrying copy...");
			$sendfile_available=0;
			$result=copyFile($source,$target,%options);
		} else {
			$result=1;
		}
	} elsif ( $mode eq 'cp' ) {
		$result=iaexec(["/bin/cp",$source,"$target$newornot"],'exec');
	} else {
		out('e',"copyFile(): Mode '$mode' is unknown. Internal error.");
		return 0;
	}
	my $success=1;
	if ( $somesums ) {
		my $checksum;
		if ( $options{trustoldchecksum} && defined($oldchecksum) ) {
			$checksum=$oldchecksum;
		} else {
			# We neither should not trust the old checksum or there is no old one
			$checksum=readFile($tempfile,format => 'string');
			unlink($tempfile);
			if ( $checksum =~ /^([a-f0-9]+)/ ) {
				out('dd',"Determined checksum '$1' while copying '$source'");
				$checksum=$1;
			} else {
				out('e',"Unexpected checksum result after copying file '$source': $checksum");
				$checksum=undef;
			}
			if ( defined($oldchecksum) && $oldchecksum ne $checksum ) {
				out('e',"Checksum failed during copy of '$source' to '$target'");
				out('i',"Old:$oldchecksum , New:$checksum");
				unlink("$target.new");
				return 0;
			}
		}
		if ( defined($checksum) ) {
			if ( defined($options{checksums}) ) {
				$options{checksums}->{$1}=$checksum;
			}
			if ( defined($options{checksum}) ) {
				${$options{checksum}}=$checksum;
			}
			if ( defined($options{checksumfile}) ) {
				my $checksumfile="$targetfolder/${digesttype}sums.txt";
				my $updatefile=sub {
					my $newsums=readSums($checksumfile);
					$newsums->{$targetdentry}=$checksum;
					return writeSums($checksumfile,$newsums);
				};
				$success &= callFileLocked("$checksumfile.lock",$updatefile);
			}
		}
	}
	if ( ! $targetisblock ) {
		my @stat=CORE::stat($source);
		if ( ! @stat ) {
			out('w',"Unable to read mtime from '$source'");
		} else {
			if ( ! utime($stat[9],$stat[9],"$target$newornot") ) {
				out('w',"Unable to set mtime on '$target$newornot'");
			}
		}
	}
	if ( $newornot ne '' ) {
		return rename("$target$newornot",$target);
	} else {
		return 1;
	}
}

=item $size=getFsObjectSize($filename,%options);

Returns the size of the given filesystem object I<$filename>.
Supported objects are regular files and block devices. I<$size>
is the apparent size - not necessarily the storage used on the
filesystem (-> sparse, incompletely used blocks, reflink, ...).



I<$size> == B<undef>, if I<$filename> doesn't exist, cannot be accessed
or the object type is not supported (-> pipe, chardev, directory, ...).

=cut

sub getFsObjectSize {
	my ($filename,%options)=@_;
	# Wierd OO-stat interface
	my $stat=stat($filename);
	if ( ! $stat ) {
		out('e',"Error stat()int '$filename': $!");
		return undef;
	}
	if ( ($stat->mode & S_IFMT) == S_IFDIR ) {
		my $sum=0;
		my $addFileSize=sub {
			if ( -f $_ ) {
				$sum += -s $_;
			}
		};
		out('dd',"Determining filetree size of '$filename'");
		find($addFileSize,$filename);
		return $sum;
	} elsif ( ($stat->mode & S_IFMT) == S_IFBLK ) {
		my $fh;
		if ( ! open($fh,"< $filename") ) {
			out('e',"Unable to open '$filename' for reading to determine size.");
			return undef;
		}
		seek($fh,0,2);
		my $size=tell($fh);
		close($fh);
		return $size;
	} elsif ( ($stat->mode & S_IFMT) == S_IFREG ) {
		return $stat->size;
	} else {
		out('e',"Invalid filesystem object '$filename' - not a file or blockdev.");
		return undef;
	}
}

=item $result=callFileLocked($lockfile,\&sub,%options)

Runs I<$result>=I<&sub()> while making sure to have an
exclusive lock on file I<$lockfile>.

Possible I<%options> are:

=over 4

=item B<noblock> => B<1>

Do not block, return lock failure instead.

=item B<lockerrorresult>

This is returned instead of B<undef> upon locking error if given.

=item B<openerrorresult>

This is returned opon lock file access error if given.

=item B<test> => B<1>

No error message output.

=back

B<undef> is returned if an exclusive lock cannot be ensured
or I<$options{lockerrorresult}> if given. Otherwise the
result of I<&sub> is returned.

=cut

sub callFileLocked {
	my ($lockfile,$sub,%options)=@_;
	my $fh;
	if ( ! open($fh,">> $lockfile") ) {
		if ( ! $options{test} ) {
			out('e',"Unable to open lock file '$lockfile'.");
		}
		if ( exists($options{openerrorresult}) ) {
			return $options{openerrorresult};
		}
		return undef;
	}
	my $lockoptions=LOCK_EX;
	if ( $options{noblock} ) {
		$lockoptions |= LOCK_NB;
	}
	if ( ! flock($fh,$lockoptions) ) {
		if ( ! $options{test} ) {
			out('e',"Unable to lock file '$lockfile'.");
		}
		if ( exists($options{lockerrorresult}) ) {
			return $options{lockerrorresult};
		}
		close($fh);
		return undef;
	}
	my $result=&{$sub};
	flock($fh,LOCK_UN);close($fh);
	return $result;
}

=back

=cut

sub prepareFtests {
	my %options=@_;
	if ( $options{iaexec} ) {
		$fake_iaexec=$options{iaexec};
		if ( $options{iaexec_prefix} ) {
			$fake_iaexec_prefix=$options{iaexec_prefix};
		}
	}
	return 1;
}

sub dryrunVerbose {
	my $dryrun=$_[0];
	my @options=();
	if ( $dryrun ) {
		push(@options,"-n");
	}
	if ( $Fb::Out::VERBOSE > 0 ) {
		push(@options,"-" . ('v' x $Fb::Out::VERBOSE));
	}
	if ( $Fb::Out::HIDDENDEBUG ) {
		push(@options,"--hidden-debug=" . int($Fb::Out::HIDDENDEBUG));
	}
	return @options;
}


1;
